# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
import json

from pyomo.common.collections import Bunch
from pyomo.common.dependencies import yaml, yaml_available, yaml_load_args
from pyomo.dataportal.factory import DataManagerFactory


def detuplize(d, sort=False):
    # print("detuplize %s" % str(d))
    if type(d) in (list, tuple, set):
        ans = []
        for item in d:
            if type(item) in (list, tuple, set):
                ans.append(list(item))
            else:
                ans.append(item)
        if sort:
            return sorted(ans)
        return ans
    elif None in d:
        return d[None]
    else:
        #
        # De-tuplize keys via list of key/value pairs
        #
        ans = []
        for k, v in d.items():
            if type(k) is tuple:
                ans.append({'index': list(k), 'value': v})
            else:
                ans.append({'index': k, 'value': v})
        if sort:
            return sorted(ans, key=lambda x: x['value'])
        return ans


def tuplize(d):
    # print("tuplize %s" % str(d))
    if type(d) is list and len(d) > 0 and not type(d[0]) is dict:
        ans = []
        for val in d:
            if type(val) is list:
                item = []
                for v in val:
                    try:
                        item.append(int(v))
                    except:
                        try:
                            item.append(float(v))
                        except:
                            item.append(v)
                item = tuple(item)
            else:
                try:
                    item = int(val)
                except:
                    try:
                        item = float(val)
                    except:
                        item = val
            ans.append(item)
        return ans
    #
    elif type(d) is list:
        ret = {}
        for val in d:
            if type(val['index']) is list:
                index = tuple(val['index'])
            else:
                index = val['index']
            ret[index] = val['value']
        return ret
    #
    elif type(d) is dict:
        return d
    else:
        return {None: d}


@DataManagerFactory.register("json", "JSON file interface")
class JSONDictionary:
    def __init__(self):
        self._info = {}
        self.options = Bunch()

    def available(self):
        return True

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")

    def close(self):
        pass

    def read(self):
        """
        This function loads data from a JSON file and tuplizes the nested
        dictionaries and lists of lists.
        """
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        INPUT = open(self.filename, 'r')
        jdata = json.load(INPUT)
        INPUT.close()
        if jdata is None or len(jdata) == 0:
            raise IOError("Empty JSON data file")
        self._info = {}
        for k, v in jdata.items():
            self._info[k] = tuplize(v)

    def write(self, data):
        """
        This function creates a JSON file for the specified data.
        """
        with open(self.filename, 'w') as OUTPUT:
            jdata = {}
            if self.options.data is None:
                for k, v in data.items():
                    jdata[k] = detuplize(v)
            elif type(self.options.data) in (list, tuple):
                for k in self.options.data:
                    jdata[k] = detuplize(data[k], sort=self.options.sort)
            else:
                k = self.options.data
                jdata[k] = detuplize(data[k])
            json.dump(jdata, OUTPUT)

    def process(self, model, data, default):
        """
        Set the data for the selected components
        """
        if not self.options.namespace in data:
            data[self.options.namespace] = {}
        #
        try:
            if self.options.data is None:
                for key in self._info:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            elif type(self.options.data) in (list, tuple):
                for key in self.options.data:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            else:
                key = self.options.data
                self._set_data(data, self.options.namespace, key, self._info[key])
        except KeyError:
            raise IOError(
                "Data value for '%s' is not available in JSON file '%s'"
                % (key, self.filename)
            )

    def _set_data(self, data, namespace, name, value):
        if type(value) is dict:
            data[namespace][name] = value
        else:
            data[namespace][name] = {None: value}

    def clear(self):
        self._info = {}


@DataManagerFactory.register("yaml", "YAML file interface")
class YamlDictionary:
    def __init__(self):
        self._info = {}
        self.options = Bunch()

    def available(self):
        return yaml_available

    def requirements(self):
        return "pyyaml"

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")

    def close(self):
        pass

    def read(self):
        """
        This function loads data from a YAML file and tuplizes the nested
        dictionaries and lists of lists.
        """
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)
        INPUT = open(self.filename, 'r')
        jdata = yaml.load(INPUT, **yaml_load_args)
        INPUT.close()
        if jdata is None:
            raise IOError("Empty YAML file")
        self._info = {}
        for k, v in jdata.items():
            self._info[k] = tuplize(v)

    def write(self, data):
        """
        This function creates a YAML file for the specified data.
        """
        with open(self.filename, 'w') as OUTPUT:
            jdata = {}
            if self.options.data is None:
                for k, v in data.items():
                    jdata[k] = detuplize(v)
            elif type(self.options.data) in (list, tuple):
                for k in self.options.data:
                    jdata[k] = detuplize(data[k], sort=self.options.sort)
            else:
                k = self.options.data
                jdata[k] = detuplize(data[k])
            yaml.dump(jdata, OUTPUT)

    def process(self, model, data, default):
        """
        Set the data for the selected components
        """
        if not self.options.namespace in data:
            data[self.options.namespace] = {}
        #
        try:
            if self.options.data is None:
                for key in self._info:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            elif type(self.options.data) in (list, tuple):
                for key in self.options.data:
                    self._set_data(data, self.options.namespace, key, self._info[key])
            else:
                key = self.options.data
                self._set_data(data, self.options.namespace, key, self._info[key])
        except KeyError:
            raise IOError(
                "Data value for '%s' is not available in YAML file '%s'"
                % (key, self.filename)
            )

    def _set_data(self, data, namespace, name, value):
        if type(value) is dict:
            data[namespace][name] = value
        else:
            data[namespace][name] = {None: value}

    def clear(self):
        self._info = {}



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo

model = pyo.ConcreteModel()
model.y = pyo.Var(bounds=(-10, 10), initialize=0.5)
model.x = pyo.Var(bounds=(-5, 5), initialize=0.5)


@model.ConstraintList()
def c(m):
    yield m.y >= m.x - 2
    yield m.y >= -m.x
    yield m.y <= m.x
    yield m.y <= 2 - m.x


model.obj = pyo.Objective(expr=model.y, sense=pyo.minimize)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import MutableMapping
from pyomo.contrib.mpc.data.dynamic_data_base import _is_iterable, _DynamicDataBase
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.series_data import TimeSeriesData
from pyomo.contrib.mpc.data.interval_data import IntervalData
from pyomo.contrib.mpc.data.find_nearest_index import find_nearest_interval_index


def _process_to_dynamic_data(data, time_set=None):
    """Processes a user's data to convert it to the appropriate type
    of dynamic data

    Mappings are converted to ScalarData, and length-two tuples are converted
    to TimeSeriesData or IntervalData, depending on the contents of the
    second item (the list of time points or intervals).

    Arguments
    ---------
    data: Dict, ComponentMap, or Tuple
        Data to convert to either ScalarData, TimeSeriesData, or
        IntervalData, depending on type.

    Returns
    -------
    ScalarData, TimeSeriesData, or IntervalData

    """
    if isinstance(data, _DynamicDataBase):
        return data
    if isinstance(data, MutableMapping):
        return ScalarData(data, time_set=time_set)
    elif isinstance(data, tuple):
        if len(data) != 2:
            raise TypeError(
                "_process_to_dynamic_data only accepts a mapping or a"
                " tuple of length two. Got tuple of length %s" % len(data)
            )
        if not isinstance(data[0], MutableMapping):
            raise TypeError(
                "First entry of data tuple must be instance of MutableMapping,"
                "e.g. dict or ComponentMap. Got %s" % type(data[0])
            )
        elif len(data[1]) == 0:
            raise ValueError(
                "Time sequence provided in data tuple is empty."
                " Cannot infer whether this is a list of points or intervals."
            )
        elif all(not _is_iterable(item) for item in data[1]):
            return TimeSeriesData(*data)
        elif all(_is_iterable(item) and len(item) == 2 for item in data[1]):
            return IntervalData(*data)
        else:
            raise TypeError(
                "Second entry of data tuple must be a non-empty iterable of"
                " scalars (time points) or length-two tuples (intervals)."
                " Got %s" % str(data[1])
            )


def interval_to_series(
    data, time_points=None, tolerance=0.0, use_left_endpoints=False, prefer_left=True
):
    """
    Arguments
    ---------
    data: IntervalData
        Data to convert to a TimeSeriesData object
    time_points: Iterable (optional)
        Points at which time series will be defined. Values are taken
        from the interval in which each point lives. The default is to
        use the right endpoint of each interval.
    tolerance: Float (optional)
        Tolerance within which time points are considered equal.
        Default is zero.
    use_left_endpoints: Bool (optional)
        Whether the left endpoints should be used in the case when
        time_points is not provided. Default is False, meaning that
        the right interval endpoints will be used. Should not be set
        if time points are provided.
    prefer_left: Bool (optional)
        If time_points is provided, and a time point is equal (within
        tolerance) to a boundary between two intervals, this flag
        controls which interval is used.

    Returns
    -------
    TimeSeriesData

    """
    if time_points is None:
        # TODO: Should first or last data points of first or last
        # intervals be included?
        if use_left_endpoints:
            time_points = [t for t, _ in data.get_intervals()]
        else:
            time_points = [t for _, t in data.get_intervals()]
        series_data = data.get_data()
        # TODO: Should TimeSeriesData be constructed with the original time set?
        return TimeSeriesData(series_data, time_points)
    if use_left_endpoints:
        raise RuntimeError("Cannot provide time_points with use_left_endpoints=True")

    intervals = data.get_intervals()
    data_dict = data.get_data()
    # NOTE: This implementation is O(len(time_points)*log(len(intervals))).
    # Could potentially do better with an O(len(time_points) + len(intervals))
    # implementation.
    idx_list = [
        find_nearest_interval_index(
            intervals, t, tolerance=tolerance, prefer_left=prefer_left
        )
        for t in time_points
    ]
    for i, t in enumerate(time_points):
        if idx_list[i] is None:
            raise RuntimeError(
                "Time point %s cannot be found in intervals within"
                " tolerance %s." % (t, tolerance)
            )
    new_data = {key: [vals[i] for i in idx_list] for key, vals in data_dict.items()}
    # TODO: Should TimeSeriesData be constructed with the original time set?
    return TimeSeriesData(new_data, time_points)


def series_to_interval(data, use_left_endpoints=False):
    """
    Arguments
    ---------
    data: TimeSeriesData
        Data that will be converted into an IntervalData object
    use_left_endpoints: Bool (optional)
        Flag indicating whether values on intervals should come
        from the values at the left or right endpoints of the
        intervals

    Returns
    -------
    IntervalData

    """
    time = data.get_time_points()
    data_dict = data.get_data()
    n_t = len(time)
    if n_t == 1:
        t0 = time[0]
        # TODO: Copy data dict?
        # TODO: Should we raise an error if time list has length one?
        return IntervalData(data_dict, [(t0, t0)])
    else:
        # This covers the case of n_t > 1 and n_t == 0
        new_data = {}
        intervals = [(time[i - 1], time[i]) for i in range(1, n_t)]
        for key, values in data_dict.items():
            interval_values = [
                values[i - 1] if use_left_endpoints else values[i]
                for i in range(1, n_t)
            ]
            new_data[key] = interval_values
        return IntervalData(new_data, intervals)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise.tests import models
import pyomo.contrib.piecewise.tests.common_tests as ct
from pyomo.core.base import TransformationFactory
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)
from pyomo.gdp import Disjunct, Disjunction
from pyomo.environ import Constraint, SolverFactory, Var


class TestTransformPiecewiseModelToReducedInnerRepnGDP(unittest.TestCase):
    def check_disjunct(self, d, not_pts):
        self.assertEqual(len(d.component_map(Constraint)), 1)
        # just the indicator_var
        self.assertEqual(len(d.component_map(Var)), 1)
        self.assertIsInstance(d.lambdas_zero_for_other_simplices, Constraint)
        self.assertEqual(len(d.lambdas_zero_for_other_simplices), len(not_pts))
        transBlock = d.parent_block()
        for i, cons in zip(not_pts, d.lambdas_zero_for_other_simplices.values()):
            assertExpressionsEqual(self, cons.expr, transBlock.lambdas[i] <= 0)

    def check_log_trans_block_structure(self, transBlock):
        m = transBlock.model()
        # One (indexed) disjunct
        self.assertEqual(len(transBlock.component_map(Disjunct)), 1)
        # One disjunction
        self.assertEqual(len(transBlock.component_map(Disjunction)), 1)
        # substitute Var and lambdas:
        self.assertEqual(len(transBlock.component_map(Var)), 2)
        # The 'z' var (that we will substitute in for the function being
        # approximated) is here:
        self.assertIsInstance(transBlock.substitute_var, Var)

        self.assertIsInstance(transBlock.lambdas, Var)
        self.assertEqual(len(transBlock.lambdas), 4)
        for lamb in transBlock.lambdas.values():
            self.assertEqual(lamb.lb, 0)
            self.assertEqual(lamb.ub, 1)
        self.assertIsInstance(transBlock.convex_combo, Constraint)
        assertExpressionsEqual(
            self,
            transBlock.convex_combo.expr,
            transBlock.lambdas[0]
            + transBlock.lambdas[1]
            + transBlock.lambdas[2]
            + transBlock.lambdas[3]
            == 1,
        )
        self.assertIsInstance(transBlock.linear_combo, Constraint)
        self.assertEqual(len(transBlock.linear_combo), 1)
        pts = m.pw_log._points
        assertExpressionsEqual(
            self,
            transBlock.linear_combo[0].expr,
            m.x
            == pts[0][0] * transBlock.lambdas[0]
            + pts[1][0] * transBlock.lambdas[1]
            + pts[2][0] * transBlock.lambdas[2]
            + pts[3][0] * transBlock.lambdas[3],
        )

        self.assertIsInstance(transBlock.linear_func, Constraint)
        self.assertEqual(len(transBlock.linear_func), 1)
        assertExpressionsEqual(
            self,
            transBlock.linear_func.expr,
            transBlock.lambdas[0] * m.f1(1)
            + transBlock.lambdas[1] * m.f1(3)
            + transBlock.lambdas[2] * m.f2(6)
            + transBlock.lambdas[3] * m.f3(10)
            == transBlock.substitute_var,
            places=7,
        )

    def check_paraboloid_trans_block_structure(self, transBlock):
        m = transBlock.model()
        # One (indexed) disjunct
        self.assertEqual(len(transBlock.component_map(Disjunct)), 1)
        # One disjunction
        self.assertEqual(len(transBlock.component_map(Disjunction)), 1)
        # substitute Var and lambdas:
        self.assertEqual(len(transBlock.component_map(Var)), 2)
        # 3 constraints: The convexity one, the x-is-a-linear-combo of extreme
        # points one, and the
        # z-is-a-linear-combo-of-pw-linear-function-values-at-extreme-ppoints
        # one:
        self.assertEqual(len(transBlock.component_map(Constraint)), 3)

        # The 'z' var (that we will substitute in for the function being
        # approximated) is here:
        self.assertIsInstance(transBlock.substitute_var, Var)

        self.assertIsInstance(transBlock.lambdas, Var)
        self.assertEqual(len(transBlock.lambdas), 6)
        for lamb in transBlock.lambdas.values():
            self.assertEqual(lamb.lb, 0)
            self.assertEqual(lamb.ub, 1)
        self.assertIsInstance(transBlock.convex_combo, Constraint)
        assertExpressionsEqual(
            self,
            transBlock.convex_combo.expr,
            transBlock.lambdas[0]
            + transBlock.lambdas[1]
            + transBlock.lambdas[2]
            + transBlock.lambdas[3]
            + transBlock.lambdas[4]
            + transBlock.lambdas[5]
            == 1,
        )
        self.assertIsInstance(transBlock.linear_combo, Constraint)
        self.assertEqual(len(transBlock.linear_combo), 2)
        pts = m.pw_paraboloid._points
        assertExpressionsEqual(
            self,
            transBlock.linear_combo[0].expr,
            m.x1
            == pts[0][0] * transBlock.lambdas[0]
            + pts[1][0] * transBlock.lambdas[1]
            + pts[2][0] * transBlock.lambdas[2]
            + pts[3][0] * transBlock.lambdas[3]
            + pts[4][0] * transBlock.lambdas[4]
            + pts[5][0] * transBlock.lambdas[5],
        )
        assertExpressionsEqual(
            self,
            transBlock.linear_combo[1].expr,
            m.x2
            == pts[0][1] * transBlock.lambdas[0]
            + pts[1][1] * transBlock.lambdas[1]
            + pts[2][1] * transBlock.lambdas[2]
            + pts[3][1] * transBlock.lambdas[3]
            + pts[4][1] * transBlock.lambdas[4]
            + pts[5][1] * transBlock.lambdas[5],
        )

        self.assertIsInstance(transBlock.linear_func, Constraint)
        self.assertEqual(len(transBlock.linear_func), 1)
        assertExpressionsEqual(
            self,
            transBlock.linear_func.expr,
            transBlock.lambdas[0] * m.g1(0, 1)
            + transBlock.lambdas[1] * m.g1(0, 4)
            + transBlock.lambdas[2] * m.g1(3, 4)
            + transBlock.lambdas[3] * m.g1(3, 1)
            + transBlock.lambdas[4] * m.g2(3, 7)
            + transBlock.lambdas[5] * m.g2(0, 7)
            == transBlock.substitute_var,
        )

    def check_pw_log(self, m):
        ##
        # Check the transformation of the approximation of log(x)
        ##
        z = m.pw_log.get_transformation_var(m.log_expr)
        self.assertIsInstance(z, Var)
        # Now we can use those Vars to check on what the transformation created
        log_block = z.parent_block()
        self.check_log_trans_block_structure(log_block)

        # Check that all of the Disjuncts have what they should
        self.assertEqual(len(log_block.disjuncts), 3)
        disjuncts_dict = {
            # disjunct : [extreme points *not* in corresponding x domain]
            log_block.disjuncts[0]: (2, 3),
            log_block.disjuncts[1]: (0, 3),
            log_block.disjuncts[2]: (0, 1),
        }
        for d, not_pts in disjuncts_dict.items():
            self.check_disjunct(d, not_pts)

        # Check the Disjunction
        self.assertIsInstance(log_block.pick_a_piece, Disjunction)
        self.assertEqual(len(log_block.pick_a_piece.disjuncts), 3)
        for i in range(2):
            self.assertIs(log_block.pick_a_piece.disjuncts[i], log_block.disjuncts[i])

        # And check the substitute Var is in the objective now.
        self.assertIs(m.obj.expr.expr, log_block.substitute_var)

    def check_pw_paraboloid(self, m):
        ##
        # Check the approximation of the transformation of the paraboloid
        ##
        z = m.pw_paraboloid.get_transformation_var(m.paraboloid_expr)
        self.assertIsInstance(z, Var)
        paraboloid_block = z.parent_block()
        self.check_paraboloid_trans_block_structure(paraboloid_block)

        self.assertEqual(len(paraboloid_block.disjuncts), 4)
        disjuncts_dict = {
            # disjunct : [extreme points *not* in corresponding (x1, x2) domain]
            paraboloid_block.disjuncts[0]: [3, 4, 5],
            paraboloid_block.disjuncts[1]: [1, 4, 5],
            paraboloid_block.disjuncts[2]: [0, 1, 3],
            paraboloid_block.disjuncts[3]: [0, 3, 4],
        }
        for d, not_pts in disjuncts_dict.items():
            self.check_disjunct(d, not_pts)

        # Check the Disjunction
        self.assertIsInstance(paraboloid_block.pick_a_piece, Disjunction)
        self.assertEqual(len(paraboloid_block.pick_a_piece.disjuncts), 4)
        for i in range(3):
            self.assertIs(
                paraboloid_block.pick_a_piece.disjuncts[i],
                paraboloid_block.disjuncts[i],
            )

        # And check the substitute Var is in the objective now.
        self.assertIs(m.indexed_c[0].body.args[0].expr, paraboloid_block.substitute_var)

    def test_transformation_do_not_descend(self):
        ct.check_transformation_do_not_descend(
            self, 'contrib.piecewise.reduced_inner_repn_gdp'
        )

    def test_transformation_PiecewiseLinearFunction_targets(self):
        ct.check_transformation_PiecewiseLinearFunction_targets(
            self, 'contrib.piecewise.reduced_inner_repn_gdp'
        )

    def test_descend_into_expressions(self):
        ct.check_descend_into_expressions(
            self, 'contrib.piecewise.reduced_inner_repn_gdp'
        )

    def test_descend_into_expressions_constraint_target(self):
        ct.check_descend_into_expressions_constraint_target(
            self, 'contrib.piecewise.reduced_inner_repn_gdp'
        )

    def test_descend_into_expressions_objective_target(self):
        ct.check_descend_into_expressions_objective_target(
            self, 'contrib.piecewise.reduced_inner_repn_gdp'
        )

    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_convex_combo_model(self):
        m = models.make_log_x_model()
        TransformationFactory('contrib.piecewise.convex_combination').apply_to(m)
        SolverFactory('gurobi').solve(m)

        ct.check_log_x_model_soln(self, m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from abc import ABCMeta, abstractmethod
from pyomo.contrib.pynumero.interfaces import pyomo_nlp, ampl_nlp
from pyomo.contrib.pynumero.sparse import BlockMatrix, BlockVector
import numpy as np
import scipy.sparse
from pyomo.common.timing import HierarchicalTimer


class BaseInteriorPointInterface(object, metaclass=ABCMeta):
    @abstractmethod
    def n_primals(self):
        pass

    @abstractmethod
    def nnz_hessian_lag(self):
        pass

    @abstractmethod
    def primals_lb(self):
        pass

    @abstractmethod
    def primals_ub(self):
        pass

    @abstractmethod
    def init_primals(self):
        pass

    @abstractmethod
    def set_primals(self, primals):
        pass

    @abstractmethod
    def get_primals(self):
        pass

    @abstractmethod
    def get_obj_factor(self):
        pass

    @abstractmethod
    def set_obj_factor(self, obj_factor):
        pass

    @abstractmethod
    def evaluate_objective(self):
        pass

    @abstractmethod
    def evaluate_grad_objective(self):
        pass

    @abstractmethod
    def n_eq_constraints(self):
        pass

    @abstractmethod
    def n_ineq_constraints(self):
        pass

    @abstractmethod
    def nnz_jacobian_eq(self):
        pass

    @abstractmethod
    def nnz_jacobian_ineq(self):
        pass

    @abstractmethod
    def ineq_lb(self):
        pass

    @abstractmethod
    def ineq_ub(self):
        pass

    @abstractmethod
    def init_duals_eq(self):
        pass

    @abstractmethod
    def init_duals_ineq(self):
        pass

    @abstractmethod
    def set_duals_eq(self, duals_eq):
        pass

    @abstractmethod
    def set_duals_ineq(self, duals_ineq):
        pass

    @abstractmethod
    def get_duals_eq(self):
        pass

    @abstractmethod
    def get_duals_ineq(self):
        pass

    @abstractmethod
    def evaluate_eq_constraints(self):
        pass

    @abstractmethod
    def evaluate_ineq_constraints(self):
        pass

    @abstractmethod
    def evaluate_jacobian_eq(self):
        pass

    @abstractmethod
    def evaluate_jacobian_ineq(self):
        pass

    @abstractmethod
    def init_slacks(self):
        pass

    @abstractmethod
    def init_duals_primals_lb(self):
        pass

    @abstractmethod
    def init_duals_primals_ub(self):
        pass

    @abstractmethod
    def init_duals_slacks_lb(self):
        pass

    @abstractmethod
    def init_duals_slacks_ub(self):
        pass

    @abstractmethod
    def set_slacks(self, slacks):
        pass

    @abstractmethod
    def set_duals_primals_lb(self, duals):
        pass

    @abstractmethod
    def set_duals_primals_ub(self, duals):
        pass

    @abstractmethod
    def set_duals_slacks_lb(self, duals):
        pass

    @abstractmethod
    def set_duals_slacks_ub(self, duals):
        pass

    @abstractmethod
    def get_slacks(self):
        pass

    @abstractmethod
    def get_duals_primals_lb(self):
        pass

    @abstractmethod
    def get_duals_primals_ub(self):
        pass

    @abstractmethod
    def get_duals_slacks_lb(self):
        pass

    @abstractmethod
    def get_duals_slacks_ub(self):
        pass

    @abstractmethod
    def set_barrier_parameter(self, barrier):
        pass

    @abstractmethod
    def evaluate_primal_dual_kkt_matrix(self, timer=None):
        pass

    @abstractmethod
    def evaluate_primal_dual_kkt_rhs(self, timer=None):
        pass

    @abstractmethod
    def set_primal_dual_kkt_solution(self, sol):
        pass

    @abstractmethod
    def get_delta_primals(self):
        pass

    @abstractmethod
    def get_delta_slacks(self):
        pass

    @abstractmethod
    def get_delta_duals_eq(self):
        pass

    @abstractmethod
    def get_delta_duals_ineq(self):
        pass

    @abstractmethod
    def get_delta_duals_primals_lb(self):
        pass

    @abstractmethod
    def get_delta_duals_primals_ub(self):
        pass

    @abstractmethod
    def get_delta_duals_slacks_lb(self):
        pass

    @abstractmethod
    def get_delta_duals_slacks_ub(self):
        pass

    def regularize_equality_gradient(self, kkt, coef, copy_kkt=True):
        raise RuntimeError(
            'Equality gradient regularization is necessary but no '
            'function has been implemented for doing so.'
        )

    def regularize_hessian(self, kkt, coef, copy_kkt=True):
        raise RuntimeError(
            'Hessian of Lagrangian regularization is necessary but no '
            'function has been implemented for doing so.'
        )


class InteriorPointInterface(BaseInteriorPointInterface):
    def __init__(self, pyomo_model):
        if type(pyomo_model) is str:
            # Assume argument is the name of an nl file
            self._nlp = ampl_nlp.AmplNLP(pyomo_model)
        else:
            self._nlp = pyomo_nlp.PyomoNLP(pyomo_model)
        self._slacks = self.init_slacks()

        # set the init_duals_primals_lb/ub from ipopt_zL_out, ipopt_zU_out if available
        # need to compress them as well and initialize the duals_primals_lb/ub
        self._init_duals_primals_lb, self._init_duals_primals_ub = (
            self._get_full_duals_primals_bounds()
        )
        self._init_duals_primals_lb[np.isneginf(self._nlp.primals_lb())] = 0
        self._init_duals_primals_ub[np.isinf(self._nlp.primals_ub())] = 0
        self._duals_primals_lb = self._init_duals_primals_lb.copy()
        self._duals_primals_ub = self._init_duals_primals_ub.copy()

        # set the init_duals_slacks_lb/ub from the init_duals_ineq
        # need to be compressed and set according to their sign
        # (-) value indicates it the upper is active, while (+) indicates
        # that lower is active
        self._init_duals_slacks_lb = self._nlp.init_duals_ineq().copy()
        self._init_duals_slacks_lb[self._init_duals_slacks_lb < 0] = 0
        self._init_duals_slacks_ub = self._nlp.init_duals_ineq().copy()
        self._init_duals_slacks_ub[self._init_duals_slacks_ub > 0] = 0
        self._init_duals_slacks_ub *= -1.0

        self._duals_slacks_lb = self._init_duals_slacks_lb.copy()
        self._duals_slacks_ub = self._init_duals_slacks_ub.copy()

        self._delta_primals = None
        self._delta_slacks = None
        self._delta_duals_eq = None
        self._delta_duals_ineq = None
        self._barrier = None

    def n_primals(self):
        return self._nlp.n_primals()

    def nnz_hessian_lag(self):
        return self._nlp.nnz_hessian_lag()

    def set_obj_factor(self, obj_factor):
        self._nlp.set_obj_factor(obj_factor)

    def get_obj_factor(self):
        return self._nlp.get_obj_factor()

    def n_eq_constraints(self):
        return self._nlp.n_eq_constraints()

    def n_ineq_constraints(self):
        return self._nlp.n_ineq_constraints()

    def nnz_jacobian_eq(self):
        return self._nlp.nnz_jacobian_eq()

    def nnz_jacobian_ineq(self):
        return self._nlp.nnz_jacobian_ineq()

    def init_primals(self):
        primals = self._nlp.init_primals()
        return primals

    def init_slacks(self):
        slacks = self._nlp.evaluate_ineq_constraints()
        return slacks

    def init_duals_eq(self):
        return self._nlp.init_duals_eq()

    def init_duals_ineq(self):
        return self._nlp.init_duals_ineq()

    def init_duals_primals_lb(self):
        return self._init_duals_primals_lb

    def init_duals_primals_ub(self):
        return self._init_duals_primals_ub

    def init_duals_slacks_lb(self):
        return self._init_duals_slacks_lb

    def init_duals_slacks_ub(self):
        return self._init_duals_slacks_ub

    def set_primals(self, primals):
        self._nlp.set_primals(primals)

    def set_slacks(self, slacks):
        self._slacks = slacks

    def set_duals_eq(self, duals):
        self._nlp.set_duals_eq(duals)

    def set_duals_ineq(self, duals):
        self._nlp.set_duals_ineq(duals)

    def set_duals_primals_lb(self, duals):
        self._duals_primals_lb = duals

    def set_duals_primals_ub(self, duals):
        self._duals_primals_ub = duals

    def set_duals_slacks_lb(self, duals):
        self._duals_slacks_lb = duals

    def set_duals_slacks_ub(self, duals):
        self._duals_slacks_ub = duals

    def get_primals(self):
        return self._nlp.get_primals()

    def get_slacks(self):
        return self._slacks

    def get_duals_eq(self):
        return self._nlp.get_duals_eq()

    def get_duals_ineq(self):
        return self._nlp.get_duals_ineq()

    def get_duals_primals_lb(self):
        return self._duals_primals_lb

    def get_duals_primals_ub(self):
        return self._duals_primals_ub

    def get_duals_slacks_lb(self):
        return self._duals_slacks_lb

    def get_duals_slacks_ub(self):
        return self._duals_slacks_ub

    def primals_lb(self):
        return self._nlp.primals_lb()

    def primals_ub(self):
        return self._nlp.primals_ub()

    def ineq_lb(self):
        return self._nlp.ineq_lb()

    def ineq_ub(self):
        return self._nlp.ineq_ub()

    def set_barrier_parameter(self, barrier):
        self._barrier = barrier

    def pyomo_nlp(self):
        return self._nlp

    def evaluate_primal_dual_kkt_matrix(self, timer=None):
        if timer is None:
            timer = HierarchicalTimer()
        timer.start('eval hess')
        hessian = self._nlp.evaluate_hessian_lag()
        timer.stop('eval hess')
        timer.start('eval jac')
        jac_eq = self._nlp.evaluate_jacobian_eq()
        jac_ineq = self._nlp.evaluate_jacobian_ineq()
        timer.stop('eval jac')

        duals_primals_lb = self._duals_primals_lb
        duals_primals_ub = self._duals_primals_ub
        duals_slacks_lb = self._duals_slacks_lb
        duals_slacks_ub = self._duals_slacks_ub
        primals = self._nlp.get_primals()

        timer.start('hess block')
        data = duals_primals_lb / (
            primals - self._nlp.primals_lb()
        ) + duals_primals_ub / (self._nlp.primals_ub() - primals)
        n = self._nlp.n_primals()
        indices = np.arange(n)
        hess_block = scipy.sparse.coo_matrix((data, (indices, indices)), shape=(n, n))
        hess_block += hessian
        timer.stop('hess block')

        timer.start('slack block')
        data = duals_slacks_lb / (
            self._slacks - self._nlp.ineq_lb()
        ) + duals_slacks_ub / (self._nlp.ineq_ub() - self._slacks)
        n = self._nlp.n_ineq_constraints()
        indices = np.arange(n)
        slack_block = scipy.sparse.coo_matrix((data, (indices, indices)), shape=(n, n))
        timer.stop('slack block')

        timer.start('set block')
        kkt = BlockMatrix(4, 4)
        kkt.set_block(0, 0, hess_block)
        kkt.set_block(1, 1, slack_block)
        kkt.set_block(2, 0, jac_eq)
        kkt.set_block(0, 2, jac_eq.transpose())
        kkt.set_block(3, 0, jac_ineq)
        kkt.set_block(0, 3, jac_ineq.transpose())
        kkt.set_block(
            3, 1, -scipy.sparse.identity(self._nlp.n_ineq_constraints(), format='coo')
        )
        kkt.set_block(
            1, 3, -scipy.sparse.identity(self._nlp.n_ineq_constraints(), format='coo')
        )
        timer.stop('set block')
        return kkt

    def evaluate_primal_dual_kkt_rhs(self, timer=None):
        if timer is None:
            timer = HierarchicalTimer()
        timer.start('eval grad obj')
        grad_obj = self.get_obj_factor() * self.evaluate_grad_objective()
        timer.stop('eval grad obj')
        timer.start('eval jac')
        jac_eq = self._nlp.evaluate_jacobian_eq()
        jac_ineq = self._nlp.evaluate_jacobian_ineq()
        timer.stop('eval jac')
        timer.start('eval cons')
        eq_resid = self._nlp.evaluate_eq_constraints()
        ineq_resid = self._nlp.evaluate_ineq_constraints() - self._slacks
        timer.stop('eval cons')

        timer.start('grad_lag_primals')
        grad_lag_primals = (
            grad_obj
            + jac_eq.transpose() * self._nlp.get_duals_eq()
            + jac_ineq.transpose() * self._nlp.get_duals_ineq()
            - self._barrier / (self._nlp.get_primals() - self._nlp.primals_lb())
            + self._barrier / (self._nlp.primals_ub() - self._nlp.get_primals())
        )
        timer.stop('grad_lag_primals')

        timer.start('grad_lag_slacks')
        grad_lag_slacks = (
            -self._nlp.get_duals_ineq()
            - self._barrier / (self._slacks - self._nlp.ineq_lb())
            + self._barrier / (self._nlp.ineq_ub() - self._slacks)
        )
        timer.stop('grad_lag_slacks')

        rhs = BlockVector(4)
        rhs.set_block(0, grad_lag_primals)
        rhs.set_block(1, grad_lag_slacks)
        rhs.set_block(2, eq_resid)
        rhs.set_block(3, ineq_resid)
        rhs = -rhs
        return rhs

    def set_primal_dual_kkt_solution(self, sol):
        self._delta_primals = sol.get_block(0)
        self._delta_slacks = sol.get_block(1)
        self._delta_duals_eq = sol.get_block(2)
        self._delta_duals_ineq = sol.get_block(3)

    def get_delta_primals(self):
        return self._delta_primals

    def get_delta_slacks(self):
        return self._delta_slacks

    def get_delta_duals_eq(self):
        return self._delta_duals_eq

    def get_delta_duals_ineq(self):
        return self._delta_duals_ineq

    def get_delta_duals_primals_lb(self):
        res = (
            (self._barrier - self._duals_primals_lb * self._delta_primals)
            / (self._nlp.get_primals() - self._nlp.primals_lb())
        ) - self._duals_primals_lb
        return res

    def get_delta_duals_primals_ub(self):
        res = (
            (self._barrier + self._duals_primals_ub * self._delta_primals)
            / (self._nlp.primals_ub() - self._nlp.get_primals())
        ) - self._duals_primals_ub
        return res

    def get_delta_duals_slacks_lb(self):
        res = (
            (self._barrier - self._duals_slacks_lb * self._delta_slacks)
            / (self._slacks - self._nlp.ineq_lb())
        ) - self._duals_slacks_lb
        return res

    def get_delta_duals_slacks_ub(self):
        res = (
            (self._barrier + self._duals_slacks_ub * self._delta_slacks)
            / (self._nlp.ineq_ub() - self._slacks)
        ) - self._duals_slacks_ub
        return res

    def evaluate_objective(self):
        return self._nlp.evaluate_objective()

    def evaluate_eq_constraints(self):
        return self._nlp.evaluate_eq_constraints()

    def evaluate_ineq_constraints(self):
        return self._nlp.evaluate_ineq_constraints()

    def evaluate_grad_objective(self):
        return self._nlp.evaluate_grad_objective()

    def evaluate_jacobian_eq(self):
        return self._nlp.evaluate_jacobian_eq()

    def evaluate_jacobian_ineq(self):
        return self._nlp.evaluate_jacobian_ineq()

    def regularize_equality_gradient(self, kkt, coef, copy_kkt=True):
        # Not technically regularizing the equality gradient ...
        # Replace this with a regularize_diagonal_block function?
        # Then call with kkt matrix and the value of the perturbation?

        # Use a constant perturbation to regularize the equality constraint
        # gradient
        if copy_kkt:
            kkt = kkt.copy()
        reg_coef = coef
        ptb = reg_coef * scipy.sparse.identity(
            self._nlp.n_eq_constraints(), format='coo'
        )

        kkt.set_block(2, 2, ptb)
        return kkt

    def regularize_hessian(self, kkt, coef, copy_kkt=True):
        if copy_kkt:
            kkt = kkt.copy()

        hess = kkt.get_block(0, 0)
        ptb = coef * scipy.sparse.identity(self._nlp.n_primals(), format='coo')
        hess += ptb
        kkt.set_block(0, 0, hess)
        return kkt

    def _get_full_duals_primals_bounds(self):
        full_duals_primals_lb = None
        full_duals_primals_ub = None
        # Check in case _nlp was constructed as an AmplNLP (from an nl file)
        if hasattr(self._nlp, 'pyomo_model') and hasattr(
            self._nlp, 'get_pyomo_variables'
        ):
            pyomo_model = self._nlp.pyomo_model()
            pyomo_variables = self._nlp.get_pyomo_variables()
            if hasattr(pyomo_model, 'ipopt_zL_out'):
                zL_suffix = pyomo_model.ipopt_zL_out
                full_duals_primals_lb = np.empty(self._nlp.n_primals())
                for i, v in enumerate(pyomo_variables):
                    if v in zL_suffix:
                        full_duals_primals_lb[i] = zL_suffix[v]

            if hasattr(pyomo_model, 'ipopt_zU_out'):
                zU_suffix = pyomo_model.ipopt_zU_out
                full_duals_primals_ub = np.empty(self._nlp.n_primals())
                for i, v in enumerate(pyomo_variables):
                    if v in zU_suffix:
                        full_duals_primals_ub[i] = zU_suffix[v]

        if full_duals_primals_lb is None:
            full_duals_primals_lb = np.ones(self._nlp.n_primals())

        if full_duals_primals_ub is None:
            full_duals_primals_ub = np.ones(self._nlp.n_primals())

        return full_duals_primals_lb, full_duals_primals_ub

    def load_primals_into_pyomo_model(self):
        if not isinstance(self._nlp, pyomo_nlp.PyomoNLP):
            raise RuntimeError(
                'Can only load primals into a pyomo model if a pyomo model was used in the constructor.'
            )

        pyomo_variables = self._nlp.get_pyomo_variables()
        primals = self._nlp.get_primals()
        for i, v in enumerate(pyomo_variables):
            v.set_value(primals[i], skip_validation=True)

    def pyomo_model(self):
        return self._nlp.pyomo_model()

    def get_pyomo_variables(self):
        return self._nlp.get_pyomo_variables()

    def get_pyomo_constraints(self):
        return self._nlp.get_pyomo_constraints()

    def variable_names(self):
        return self._nlp.variable_names()

    def constraint_names(self):
        return self._nlp.constraint_names()

    def get_primal_indices(self, pyomo_variables):
        return self._nlp.get_primal_indices(pyomo_variables)

    def get_constraint_indices(self, pyomo_constraints):
        return self._nlp.get_constraint_indices(pyomo_constraints)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.cp.interval_var import IntervalVar
from pyomo.contrib.cp.scheduling_expr.scheduling_logic import (
    AlternativeExpression,
    SpanExpression,
    SynchronizeExpression,
    alternative,
    spans,
    synchronize,
)
from pyomo.contrib.cp.scheduling_expr.sequence_expressions import (
    NoOverlapExpression,
    FirstInSequenceExpression,
    LastInSequenceExpression,
    BeforeInSequenceExpression,
    PredecessorToExpression,
    no_overlap,
    predecessor_to,
    before_in_sequence,
    first_in_sequence,
    last_in_sequence,
)
from pyomo.contrib.cp.sequence_var import SequenceVar
from pyomo.environ import ConcreteModel, LogicalConstraint, Set


class TestSequenceVarExpressions(unittest.TestCase):
    def get_model(self):
        m = ConcreteModel()
        m.S = Set(initialize=range(3))
        m.i = IntervalVar(m.S, start=(0, 5))
        m.seq = SequenceVar(expr=[m.i[j] for j in m.S])

        return m

    def test_no_overlap(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=no_overlap(m.seq))
        e = m.c.expr

        self.assertIsInstance(e, NoOverlapExpression)
        self.assertEqual(e.nargs(), 1)
        self.assertEqual(len(e.args), 1)
        self.assertIs(e.args[0], m.seq)

        self.assertEqual(str(e), "no_overlap(seq)")

    def test_first_in_sequence(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=first_in_sequence(m.i[2], m.seq))
        e = m.c.expr

        self.assertIsInstance(e, FirstInSequenceExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertEqual(len(e.args), 2)
        self.assertIs(e.args[0], m.i[2])
        self.assertIs(e.args[1], m.seq)

        self.assertEqual(str(e), "first_in(i[2], seq)")

    def test_last_in_sequence(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=last_in_sequence(m.i[0], m.seq))
        e = m.c.expr

        self.assertIsInstance(e, LastInSequenceExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertEqual(len(e.args), 2)
        self.assertIs(e.args[0], m.i[0])
        self.assertIs(e.args[1], m.seq)

        self.assertEqual(str(e), "last_in(i[0], seq)")

    def test_before_in_sequence(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=before_in_sequence(m.i[1], m.i[0], m.seq))
        e = m.c.expr

        self.assertIsInstance(e, BeforeInSequenceExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertEqual(len(e.args), 3)
        self.assertIs(e.args[0], m.i[1])
        self.assertIs(e.args[1], m.i[0])
        self.assertIs(e.args[2], m.seq)

        self.assertEqual(str(e), "before_in(i[1], i[0], seq)")

    def test_predecessor_in_sequence(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=predecessor_to(m.i[0], m.i[1], m.seq))
        e = m.c.expr

        self.assertIsInstance(e, PredecessorToExpression)
        self.assertEqual(e.nargs(), 3)
        self.assertEqual(len(e.args), 3)
        self.assertIs(e.args[0], m.i[0])
        self.assertIs(e.args[1], m.i[1])
        self.assertIs(e.args[2], m.seq)

        self.assertEqual(str(e), "predecessor_to(i[0], i[1], seq)")


class TestHierarchicalSchedulingExpressions(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()

        def start_rule(m, i):
            return 2 * i

        def length_rule(m, i):
            return i

        m.iv = IntervalVar(
            [1, 2, 3], start=start_rule, length=length_rule, optional=True
        )
        m.whole_enchilada = IntervalVar()

        return m

    def check_span_expression(self, m, e):
        self.assertIsInstance(e, SpanExpression)
        self.assertEqual(e.nargs(), 4)
        self.assertEqual(len(e.args), 4)
        self.assertIs(e.args[0], m.whole_enchilada)
        for i in [1, 2, 3]:
            self.assertIs(e.args[i], m.iv[i])

        self.assertEqual(str(e), "whole_enchilada.spans(iv[1], iv[2], iv[3])")

    def test_spans(self):
        m = self.make_model()
        e = spans(m.whole_enchilada, [m.iv[i] for i in [1, 2, 3]])
        self.check_span_expression(m, e)

    def test_spans_method(self):
        m = self.make_model()
        e = m.whole_enchilada.spans(m.iv[i] for i in [1, 2, 3])
        self.check_span_expression(m, e)

    def test_alternative(self):
        m = self.make_model()
        e = alternative(m.whole_enchilada, [m.iv[i] for i in [1, 2, 3]])

        self.assertIsInstance(e, AlternativeExpression)
        self.assertEqual(e.nargs(), 4)
        self.assertEqual(len(e.args), 4)
        self.assertIs(e.args[0], m.whole_enchilada)
        for i in [1, 2, 3]:
            self.assertIs(e.args[i], m.iv[i])

        self.assertEqual(str(e), "alternative(whole_enchilada, [iv[1], iv[2], iv[3]])")

    def test_synchronize(self):
        m = self.make_model()
        e = synchronize(m.whole_enchilada, [m.iv[i] for i in [1, 2, 3]])

        self.assertIsInstance(e, SynchronizeExpression)
        self.assertEqual(e.nargs(), 4)
        self.assertEqual(len(e.args), 4)
        self.assertIs(e.args[0], m.whole_enchilada)
        for i in [1, 2, 3]:
            self.assertIs(e.args[i], m.iv[i])

        self.assertEqual(str(e), "synchronize(whole_enchilada, [iv[1], iv[2], iv[3]])")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the PyUtilib project
# Copyright (c) 2008 Sandia Corporation.
# This software is distributed under the BSD License.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
# ____________________________________________________________________________________

import shlex
from collections.abc import Mapping


class Bunch(dict):
    """A class that can be used to store a bunch of data dynamically.

    This class allows for unspecified attributes to have a default value
    of None.  This borrows the output formatting ideas from the
    ActiveState Code Container (recipe 496697).

    For historical reasons, attributes / keys are stored in the
    underlying dict unless they begin with an underscore, in which case
    they are stored as object attributes.

    """

    def __init__(self, *args, **kw):
        self._name_ = self.__class__.__name__
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Bunch() positional arguments must be strings")
            for item in shlex.split(arg):
                item = item.split('=', 1)
                if len(item) != 2:
                    raise ValueError(
                        "Bunch() positional arguments must be space separated "
                        f"strings of form 'key=value', got '{item[0]}'"
                    )

                # Historically, this used 'exec'.  That is unsafe in
                # this context (because anyone can pass arguments to a
                # Bunch).  While not strictly backwards compatible,
                # Pyomo was not using this for anything past parsing
                # None/float/int values.  We will explicitly parse those
                # values
                try:
                    val = float(item[1])
                    if int(val) == val:
                        val = int(val)
                    item[1] = val
                except:
                    if item[1].strip() == 'None':
                        item[1] = None
                self[item[0]] = item[1]
        for k, v in kw.items():
            self[k] = v

    def update(self, d):
        """
        The update is specialized for JSON-like data.  This
        recursively replaces dictionaries with Bunch objects.
        """

        def _replace_dict_in_list(lst):
            ans = []
            for v in lst:
                if type(v) is dict:
                    ans.append(Bunch())
                    ans[-1].update(v)
                elif type(v) is list:
                    ans.append(_replace_dict_in_list(v))
                else:
                    ans.append(v)
            return ans

        if isinstance(d, Mapping):
            item_iter = d.items()
        else:
            item_iter = d
        for k, v in item_iter:
            if type(v) is dict:
                self[k] = Bunch()
                self[k].update(v)
            elif type(v) is list:
                self[k] = _replace_dict_in_list(v)
            else:
                self[k] = v

    def set_name(self, name):
        self._name_ = name

    def __getitem__(self, name):
        if not isinstance(name, str):
            raise ValueError(f'Bunch keys must be str (got {type(name).__name__})')
        # Map through Python's standard getattr functionality (which
        # will resolve known attributes without hitting __getattr__)
        return getattr(self, name)

    def __setitem__(self, name, val):
        if not isinstance(name, str):
            raise ValueError(f'Bunch keys must be str (got {type(name).__name__})')
        setattr(self, name, val)

    def __delitem__(self, name):
        if not isinstance(name, str):
            raise ValueError(f'Bunch keys must be str (got {type(name).__name__})')
        delattr(self, name)

    def __getattr__(self, name):
        if name[0] == '_':
            raise AttributeError(f"Unknown attribute '{name}'")
        return self.get(name, None)

    def __setattr__(self, name, val):
        if name[0] == '_':
            super().__setattr__(name, val)
        else:
            super().__setitem__(name, val)

    def __delattr__(self, name):
        if name[0] == '_':
            super().__delattr__(name)
        else:
            super().__delitem__(name)

    def __repr__(self):
        attrs = sorted(
            "%s = %r" % (k, v) for k, v in self.items() if not k.startswith("_")
        )
        return "%s(%s)" % (self.__class__.__name__, ", ".join(attrs))

    def __str__(self, nesting=0, indent=''):
        attrs = []
        indentation = indent + "    " * nesting
        for k, v in self.items():
            text = [indentation, k, ":"]
            if isinstance(v, Bunch):
                if len(v) > 0:
                    text.append('\n')
                text.append(v.__str__(nesting + 1))
            elif isinstance(v, list):
                if len(v) == 0:
                    text.append(' []')
                else:
                    for v_ in v:
                        text.append('\n' + indentation + "-")
                        if isinstance(v_, Bunch):
                            text.append('\n' + v_.__str__(nesting + 1))
                        else:
                            text.append(" " + repr(v_))
            else:
                text.append(' ' + repr(v))
            attrs.append("".join(text))
        attrs.sort()
        return "\n".join(attrs)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


def load():
    from pyomo.repn.plugins import (
        cpxlp,
        ampl,
        baron_writer,
        mps,
        gams_writer,
        lp_writer,
        nl_writer,
        standard_form,
        parameterized_standard_form,
    )
    from pyomo.opt import WriterFactory

    # Register the "default" versions of writers that have more than one
    # implementation
    WriterFactory.register('nl', 'Generate the corresponding AMPL NL file.')(
        WriterFactory.get_class('nl_v2')
    )
    WriterFactory.register('lp', 'Generate the corresponding CPLEX LP file.')(
        WriterFactory.get_class('lp_v2')
    )
    WriterFactory.register('cpxlp', 'Generate the corresponding CPLEX LP file.')(
        WriterFactory.get_class('cpxlp_v2')
    )


def activate_writer_version(name, ver):
    """DEBUGGING TOOL to switch the "default" writer implementation"""
    from pyomo.opt import WriterFactory

    doc = WriterFactory.doc(name)
    WriterFactory.unregister(name)
    WriterFactory.register(name, doc)(WriterFactory.get_class(f'{name}_v{ver}'))


def active_writer_version(name):
    """DEBUGGING TOOL to switch the "default" writer implementation"""
    from pyomo.opt import WriterFactory

    ref = WriterFactory.get_class(name)
    ver = 1
    try:
        while 1:
            if WriterFactory.get_class(f'{name}_v{ver}') is ref:
                return ver
            ver += 1
    except KeyError:
        return None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.parmest.utils.create_ef import (
    get_objs,
    create_EF,
    find_active_objective,
    ef_nonants,
)

from pyomo.contrib.parmest.utils.ipopt_solver_wrapper import ipopt_solve_with_stats

from pyomo.contrib.parmest.utils.model_utils import convert_params_to_vars

from pyomo.contrib.parmest.utils.mpi_utils import MPIInterface, ParallelTaskManager

from pyomo.contrib.parmest.utils.scenario_tree import build_vardatalist, ScenarioNode



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys as _sys

import importlib


def _do_import(pkg_name):
    importlib.import_module(pkg_name)


#
# These packages contain plugins that need to be loaded
#
_packages = [
    'pyomo.common',
    'pyomo.core',
    'pyomo.opt',
    'pyomo.dataportal',
    'pyomo.duality',
    'pyomo.repn',
    'pyomo.neos',
    'pyomo.solvers',
    'pyomo.gdp',
    'pyomo.mpec',
    'pyomo.dae',
    'pyomo.scripting',
    'pyomo.network',
    'pyomo.contrib.ampl_function_demo',
    'pyomo.contrib.appsi',
    'pyomo.contrib.community_detection',
    'pyomo.contrib.cp',
    'pyomo.contrib.cspline_external',
    'pyomo.contrib.fme',
    'pyomo.contrib.gdp_bounds',
    'pyomo.contrib.gdpopt',
    'pyomo.contrib.gjh',
    'pyomo.contrib.gdp_bounds',
    'pyomo.contrib.aslfunctions',
    'pyomo.contrib.mcpp',
    'pyomo.contrib.mindtpy',
    'pyomo.contrib.multistart',
    'pyomo.contrib.preprocessing',
    'pyomo.contrib.pynumero',
    'pyomo.contrib.simplification',
    'pyomo.contrib.solver',
    'pyomo.contrib.trustregion',
]


def _import_packages():
    #
    # Import required packages
    #
    for _package in _packages:
        pname = _package + '.plugins'
        try:
            _do_import(pname)
        except ImportError:
            exctype, err, tb = _sys.exc_info()  # BUG?
            import traceback

            msg = (
                "pyomo.environ failed to import %s:\nOriginal %s: %s\n"
                "Traceback:\n%s"
                % (pname, exctype.__name__, err, ''.join(traceback.format_tb(tb)))
            )
            # clear local variables to remove circular references
            exctype = err = tb = None
            # TODO: Should this just log an error and re-raise the
            # original exception?
            raise ImportError(msg)

        pkg = _sys.modules[pname]
        pkg.load()


_import_packages()

#
# Expose the symbols from pyomo.core
#
from pyomo.dataportal import DataPortal
import pyomo.core.kernel
from pyomo.common.collections import ComponentMap
import pyomo.core.base.indexed_component
import pyomo.core.base.util
from pyomo.core import expr, base, kernel, plugins
from pyomo.core.base import util

from pyomo.core import (
    numvalue,
    numeric_expr,
    boolean_value,
    symbol_map,
    sympy_tools,
    taylor_series,
    visitor,
    expr_common,
    expr_errors,
    calculus,
    native_types,
    linear_expression,
    nonlinear_expression,
    land,
    lor,
    equivalent,
    exactly,
    atleast,
    atmost,
    all_different,
    count_if,
    implies,
    lnot,
    xor,
    inequality,
    log,
    log10,
    sin,
    cos,
    tan,
    cosh,
    sinh,
    tanh,
    asin,
    acos,
    atan,
    exp,
    sqrt,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    Expr_if,
    differentiate,
    taylor_series_expansion,
    SymbolMap,
    PyomoObject,
    nonpyomo_leaf_types,
    native_numeric_types,
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    polynomial_degree,
    NumericValue,
    ZeroConstant,
    as_boolean,
    BooleanConstant,
    BooleanValue,
    native_logical_values,
    minimize,
    maximize,
    PyomoOptions,
    Expression,
    CuidLabeler,
    CounterLabeler,
    NumericLabeler,
    CNameLabeler,
    TextLabeler,
    AlphaNumericTextLabeler,
    NameLabeler,
    ShortNameLabeler,
    name,
    Component,
    ComponentUID,
    BuildAction,
    BuildCheck,
    Set,
    SetOf,
    simple_set_rule,
    RangeSet,
    Param,
    Var,
    VarList,
    ScalarVar,
    BooleanVar,
    BooleanVarList,
    ScalarBooleanVar,
    logical_expr,
    simple_constraint_rule,
    simple_constraintlist_rule,
    ConstraintList,
    Constraint,
    LogicalConstraint,
    LogicalConstraintList,
    simple_objective_rule,
    simple_objectivelist_rule,
    Objective,
    ObjectiveList,
    Connector,
    SOSConstraint,
    Piecewise,
    active_export_suffix_generator,
    active_import_suffix_generator,
    Suffix,
    ExternalFunction,
    symbol_map_from_instance,
    Reference,
    Reals,
    PositiveReals,
    NonPositiveReals,
    NegativeReals,
    NonNegativeReals,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Boolean,
    Binary,
    Any,
    AnyWithNone,
    EmptySet,
    UnitInterval,
    PercentFraction,
    RealInterval,
    IntegerInterval,
    display,
    SortComponents,
    TraversalStrategy,
    Block,
    ScalarBlock,
    active_components,
    components,
    active_components_data,
    components_data,
    global_option,
    Model,
    ConcreteModel,
    AbstractModel,
    ModelComponentFactory,
    Transformation,
    TransformationFactory,
    instance2dat,
    set_options,
    RealSet,
    IntegerSet,
    BooleanSet,
    prod,
    quicksum,
    sum_product,
    dot_product,
    summation,
    sequence,
)

from pyomo.opt import (
    SolverFactory,
    SolverManagerFactory,
    UnknownSolver,
    TerminationCondition,
    SolverStatus,
    check_optimal_termination,
    assert_optimal_termination,
)
from pyomo.core.base.units_container import units, as_quantity

# These APIs are deprecated and should be removed in the near future
from pyomo.common.deprecation import relocated_module_attribute

relocated_module_attribute(
    'SimpleBlock', 'pyomo.core.base.block.SimpleBlock', version='6.0'
)
relocated_module_attribute('SimpleVar', 'pyomo.core.base.var.SimpleVar', version='6.0')
relocated_module_attribute(
    'SimpleBooleanVar', 'pyomo.core.base.boolean_var.SimpleBooleanVar', version='6.0'
)
del relocated_module_attribute



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.contrib import appsi
import pyomo.environ as pyo
from pyomo.core.base.var import ScalarVar


class TestResults(unittest.TestCase):
    def test_uninitialized(self):
        res = appsi.base.Results()
        self.assertIsNone(res.best_feasible_objective)
        self.assertIsNone(res.best_objective_bound)
        self.assertEqual(
            res.termination_condition, appsi.base.TerminationCondition.unknown
        )

        with self.assertRaisesRegex(
            RuntimeError, '.*does not currently have a valid solution.*'
        ):
            res.solution_loader.load_vars()
        with self.assertRaisesRegex(
            RuntimeError, '.*does not currently have valid duals.*'
        ):
            res.solution_loader.get_duals()
        with self.assertRaisesRegex(
            RuntimeError, '.*does not currently have valid reduced costs.*'
        ):
            res.solution_loader.get_reduced_costs()
        with self.assertRaisesRegex(
            RuntimeError, '.*does not currently have valid slacks.*'
        ):
            res.solution_loader.get_slacks()

    def test_results(self):
        m = pyo.ConcreteModel()
        m.x = ScalarVar()
        m.y = ScalarVar()
        m.c1 = pyo.Constraint(expr=m.x == 1)
        m.c2 = pyo.Constraint(expr=m.y == 2)

        primals = dict()
        primals[id(m.x)] = (m.x, 1)
        primals[id(m.y)] = (m.y, 2)
        duals = dict()
        duals[m.c1] = 3
        duals[m.c2] = 4
        rc = dict()
        rc[id(m.x)] = (m.x, 5)
        rc[id(m.y)] = (m.y, 6)
        slacks = dict()
        slacks[m.c1] = 7
        slacks[m.c2] = 8

        res = appsi.base.Results()
        res.solution_loader = appsi.base.SolutionLoader(
            primals=primals, duals=duals, slacks=slacks, reduced_costs=rc
        )

        res.solution_loader.load_vars()
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 2)

        m.x.value = None
        m.y.value = None

        res.solution_loader.load_vars([m.y])
        self.assertIsNone(m.x.value)
        self.assertAlmostEqual(m.y.value, 2)

        duals2 = res.solution_loader.get_duals()
        self.assertAlmostEqual(duals[m.c1], duals2[m.c1])
        self.assertAlmostEqual(duals[m.c2], duals2[m.c2])

        duals2 = res.solution_loader.get_duals([m.c2])
        self.assertNotIn(m.c1, duals2)
        self.assertAlmostEqual(duals[m.c2], duals2[m.c2])

        rc2 = res.solution_loader.get_reduced_costs()
        self.assertAlmostEqual(rc[id(m.x)][1], rc2[m.x])
        self.assertAlmostEqual(rc[id(m.y)][1], rc2[m.y])

        rc2 = res.solution_loader.get_reduced_costs([m.y])
        self.assertNotIn(m.x, rc2)
        self.assertAlmostEqual(rc[id(m.y)][1], rc2[m.y])

        slacks2 = res.solution_loader.get_slacks()
        self.assertAlmostEqual(slacks[m.c1], slacks2[m.c1])
        self.assertAlmostEqual(slacks[m.c2], slacks2[m.c2])

        slacks2 = res.solution_loader.get_slacks([m.c2])
        self.assertNotIn(m.c1, slacks2)
        self.assertAlmostEqual(slacks[m.c2], slacks2[m.c2])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc

from pyomo.core.kernel.homogeneous_container import IHomogeneousContainer


class TupleContainer(IHomogeneousContainer, collections.abc.Sequence):
    """
    A partial implementation of the IHomogeneousContainer
    interface that provides tuple-like storage functionality.

    Complete implementations need to set the _ctype property
    at the class level and initialize the remaining
    ICategorizedObject attributes during object creation. If
    using __slots__, a slot named "_data" must be included.

    Note that this implementation allows nested storage of
    other ICategorizedObjectContainer implementations that
    are defined with the same ctype.
    """

    __slots__ = ()
    _child_storage_delimiter_string = ""
    _child_storage_entry_string = "[%s]"

    def __init__(self, *args):
        self._data = []
        self._init(args)
        self._data = tuple(self._data)

    def _init(self, args):
        if len(args) > 0:
            if len(args) > 1:
                raise TypeError(
                    "%s expected at most 1 arguments, "
                    "got %s" % (self.__class__.__name__, len(args))
                )
            for item in args[0]:
                self._insert(len(self), item)

    def _fast_insert(self, i, item):
        item._update_parent_and_storage_key(self, i)
        self._data.insert(i, item)

    def _insert(self, i, item):
        if item.ctype is self.ctype:
            if item._parent is None:
                self._fast_insert(i, item)
                return
            # see note about allowing objects to live
            # in more than one container
            raise ValueError(
                "Invalid assignment to type %s with index %s. "
                "A parent container has already been "
                "assigned to the object being inserted: %s"
                % (self.__class__.__name__, i, item.parent.name)
            )
        else:
            raise TypeError(
                "Invalid assignment to type %s with index %s. "
                "The object being inserted has the wrong "
                "category type: %s" % (self.__class__.__name__, i, item.ctype)
            )

    #
    # Define the ICategorizedObjectContainer abstract methods
    #

    def child(self, key):
        """Get the child object associated with a given
        storage key for this container.

        Raises:
            KeyError: if the argument is not a storage key
                for any children of this container
        """
        try:
            return self.__getitem__(key)
        except (IndexError, TypeError):
            raise KeyError(str(key))

    def children(self):
        """A generator over the children of this container."""
        return self._data.__iter__()

    #
    # Define the Sequence abstract methods
    #

    def __getitem__(self, i):
        return self._data[i]

    def __len__(self):
        return self._data.__len__()

    #
    # Extend the interface to allow for equality comparison
    #
    # We want to avoid generating Pyomo expressions due to
    # comparison of values.

    # Convert both objects to a plain tuple of (type(val),
    # id(val)) tuples and compare that instead.
    def __eq__(self, other):
        if not isinstance(other, (collections.abc.Set, collections.abc.Sequence)):
            return False
        return tuple((type(val), id(val)) for val in self) == tuple(
            (type(val), id(val)) for val in other
        )

    def __ne__(self, other):
        return not (self == other)

    #
    # Override a few default implementations on Sequence
    #
    # We want to avoid generating Pyomo expressions due to
    # comparison of values.

    def __iter__(self):
        return self._data.__iter__()

    def __contains__(self, item):
        item_id = id(item)
        return any(item_id == id(_v) for _v in self._data)

    def index(self, item, start=0, stop=None):
        """S.index(value, [start, [stop]]) -> integer -- return first index of value.

        Raises ValueError if the value is not present.
        """
        if start is not None and start < 0:
            start = max(len(self) + start, 0)
        if stop is not None and stop < 0:
            stop += len(self)

        i = start
        while stop is None or i < stop:
            try:
                if self[i] is item:
                    return i
            except IndexError:
                break
            i += 1
        raise ValueError

    def count(self, item):
        'S.count(value) -> integer -- return number of occurrences of value'
        item_id = id(item)
        cnt = sum(1 for _v in self._data if id(_v) == item_id)
        assert cnt == 1
        return cnt



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import document_kwargs_from_configdict
from pyomo.contrib.gdpopt.algorithm_base_class import _GDPoptAlgorithm
from pyomo.contrib.gdpopt.config_options import (
    _add_mip_solver_configs,
    _add_nlp_solver_configs,
    _add_tolerance_configs,
    _add_oa_configs,
)
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    _get_discrete_problem_and_subproblem,
)
from pyomo.contrib.gdpopt.oa_algorithm_utils import _OAAlgorithmMixIn
from pyomo.contrib.gdpopt.cut_generation import add_no_good_cut
from pyomo.contrib.gdpopt.solve_discrete_problem import solve_MILP_discrete_problem
from pyomo.contrib.gdpopt.util import time_code
from pyomo.core import Objective
from pyomo.opt.base import SolverFactory

# ESJ: In the future, if we have a direct interface to cplex or gurobi, we
# should get the integer solutions several-at-a-time with a solution pool or
# something of the like...


@SolverFactory.register(
    'gdpopt.ric',
    doc="The RIC (relaxation with integer cuts) Generalized Disjunctive "
    "Programming (GDP) solver",
)
class GDP_RIC_Solver(_GDPoptAlgorithm, _OAAlgorithmMixIn):
    """The GDPopt (Generalized Disjunctive Programming optimizer) relaxation
    with integer cuts (RIC) solver.

    Accepts models that can include nonlinear, continuous variables and
    constraints, as well as logical conditions. For non-convex problems, RIC
    will not be exact unless the NLP subproblems are solved globally.
    """

    CONFIG = _GDPoptAlgorithm.CONFIG()
    _add_mip_solver_configs(CONFIG)
    _add_nlp_solver_configs(CONFIG, default_solver='ipopt')
    _add_tolerance_configs(CONFIG)
    _add_oa_configs(CONFIG)

    algorithm = 'RIC'

    # Override solve() to customize the docstring for this solver
    @document_kwargs_from_configdict(CONFIG, doc=_GDPoptAlgorithm.solve.__doc__)
    def solve(self, model, **kwds):
        return super().solve(model, **kwds)

    def _solve_gdp(self, original_model, config):
        logger = config.logger

        discrete_problem_util_block, subproblem_util_block = (
            _get_discrete_problem_and_subproblem(self, config)
        )
        discrete_problem = discrete_problem_util_block.parent_block()
        subproblem = subproblem_util_block.parent_block()
        discrete_problem_obj = next(
            discrete_problem.component_data_objects(
                Objective, active=True, descend_into=True
            )
        )

        self._log_header(logger)

        # main loop
        while not config.iterlim or self.iteration < config.iterlim:
            self.iteration += 1

            # solve linear discrete problem
            with time_code(self.timing, 'mip'):
                mip_feasible = solve_MILP_discrete_problem(
                    discrete_problem_util_block, self, config
                )
                self._update_bounds_after_discrete_problem_solve(
                    mip_feasible, discrete_problem_obj, logger
                )

            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

            with time_code(self.timing, 'nlp'):
                self._fix_discrete_soln_solve_subproblem_and_add_cuts(
                    discrete_problem_util_block, subproblem_util_block, config
                )

            # Add integer cut
            with time_code(self.timing, "integer cut generation"):
                add_no_good_cut(discrete_problem_util_block, config)

            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

    def _add_cuts_to_discrete_problem(
        self,
        subproblem_util_block,
        discrete_problem_util_block,
        objective_sense,
        config,
        timing,
    ):
        # Nothing to do here
        pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests for the Logic-based Branch and Bound solver plugin."""

from io import StringIO
import logging
from math import fabs
from os.path import abspath, dirname, join, normpath

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import import_file
from pyomo.common.log import LoggingIntercept
import pyomo.contrib.gdpopt.tests.common_tests as ct
from pyomo.contrib.satsolver.satsolver import z3_available
from pyomo.environ import SolverFactory, value, ConcreteModel, Var, Objective, maximize
from pyomo.gdp import Disjunction
from pyomo.opt import TerminationCondition

currdir = dirname(abspath(__file__))
exdir = normpath(join(currdir, '..', '..', '..', '..', 'examples', 'gdp'))

minlp_solver = 'baron'
minlp_args = dict()
solver_available = SolverFactory(minlp_solver).available()
license_available = (
    SolverFactory(minlp_solver).license_is_valid() if solver_available else False
)


@unittest.skipUnless(
    solver_available, "Required subsolver %s is not available" % (minlp_solver,)
)
class TestGDPopt_LBB(unittest.TestCase):
    """Tests for logic-based branch and bound."""

    def test_infeasible_GDP(self):
        """Test for infeasible GDP."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 2))
        m.d = Disjunction(expr=[[m.x**2 >= 3, m.x >= 3], [m.x**2 <= -1, m.x <= -1]])
        m.o = Objective(expr=m.x)
        result = SolverFactory('gdpopt.lbb').solve(
            m, tee=False, minlp_solver=minlp_solver, minlp_solver_args=minlp_args
        )
        self.assertEqual(
            result.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.d.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.d.disjuncts[1].indicator_var.value)

    @unittest.skipUnless(z3_available, "Z3 SAT solver is not available")
    def test_infeasible_GDP_check_sat(self):
        """Test for infeasible GDP with check_sat option True."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 2))
        m.d = Disjunction(expr=[[m.x**2 >= 3, m.x >= 3], [m.x**2 <= -1, m.x <= -1]])
        m.o = Objective(expr=m.x)
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.contrib.gdpopt', logging.INFO):
            result = SolverFactory('gdpopt.lbb').solve(
                m,
                tee=False,
                check_sat=True,
                minlp_solver=minlp_solver,
                minlp_solver_args=minlp_args,
            )
        self.assertIn(
            "Root node is not satisfiable. Problem is infeasible.",
            output.getvalue().strip(),
        )

        self.assertEqual(
            result.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.d.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.d.disjuncts[1].indicator_var.value)

    # This should work--see issue #2483
    # def test_fix_all_but_one_disjunct(self):
    #     m = ConcreteModel()
    #     m.x = Var(bounds=(0, 3))
    #     m.d = Disjunction(expr=[
    #         [m.x ** 2 >= 3, m.x >= 3],
    #         [m.x ** 2 <= -1, m.x <= -1]])
    #     m.o = Objective(expr=m.x)
    #     m.d.disjuncts[1].indicator_var.fix(False)
    #     result = SolverFactory('gdpopt.lbb').solve(
    #         m, tee=True,
    #         minlp_solver=minlp_solver,
    #         minlp_solver_args=minlp_args,
    #     )
    #     self.assertEqual(result.solver.termination_condition,
    #                      TerminationCondition.optimal)
    #     self.assertAlmostEqual(value(m.x), 3)
    #     self.assertTrue(value(m.d.disjuncts[0].indicator_var))
    #     self.assertFalse(value(m.d.disjuncts[1].indicator_var))

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    def test_LBB_8PP(self):
        """Test the logic-based branch and bound algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.lbb').solve(
            eight_process,
            tee=False,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    def test_LBB_8PP_max(self):
        """Test the logic-based branch and bound algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        obj = next(eight_process.component_data_objects(Objective, active=True))
        obj.sense = maximize
        obj.set_value(-1 * obj.expr)
        SolverFactory('gdpopt.lbb').solve(
            eight_process,
            tee=False,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        self.assertAlmostEqual(value(eight_process.profit.expr), -68, places=1)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    def test_LBB_strip_pack(self):
        """Test logic-based branch and bound with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.lbb').solve(
            strip_pack,
            tee=False,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    @unittest.pytest.mark.expensive
    def test_LBB_constrained_layout(self):
        """Test LBB with constrained layout."""
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        cons_layout = exfile.build_constrained_layout_model()
        SolverFactory('gdpopt.lbb').solve(
            cons_layout,
            tee=False,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        objective_value = value(cons_layout.min_dist_cost.expr)
        self.assertTrue(
            fabs(objective_value - 41573) <= 200,
            "Objective value of %s instead of 41573" % objective_value,
        )

    def test_LBB_ex_633_trespalacios(self):
        """Test LBB with Francisco thesis example."""
        exfile = import_file(join(exdir, 'small_lit', 'ex_633_trespalacios.py'))
        model = exfile.build_simple_nonconvex_gdp()
        SolverFactory('gdpopt.lbb').solve(
            model, tee=False, minlp_solver=minlp_solver, minlp_solver_args=minlp_args
        )
        objective_value = value(model.obj.expr)
        self.assertAlmostEqual(objective_value, 4.46, 2)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    @unittest.skipUnless(
        SolverFactory('bonmin').available(exception_flag=False),
        "Bonmin is not available",
    )
    def test_LBB_8PP_with_screening(self):
        """Test the logic-based branch and bound algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.lbb').solve(
            eight_process,
            tee=False,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
            solve_local_rnGDP=True,
            local_minlp_solver='bonmin',
            local_minlp_solver_args={},
        )
        ct.check_8PP_solution(self, eight_process, results)


@unittest.skipUnless(
    solver_available, "Required subsolver %s is not available" % (minlp_solver,)
)
@unittest.skipUnless(z3_available, "Z3 SAT solver is not available.")
class TestGDPopt_LBB_Z3(unittest.TestCase):
    """Tests for logic-based branch and bound with Z3 SAT solver integration."""

    def test_infeasible_GDP(self):
        """Test for infeasible GDP."""
        m = ConcreteModel()
        m.x = Var(bounds=(0, 2))
        m.d = Disjunction(expr=[[m.x**2 >= 3, m.x >= 3], [m.x**2 <= -1, m.x <= -1]])
        m.o = Objective(expr=m.x)
        result = SolverFactory('gdpopt.lbb').solve(
            m, tee=False, minlp_solver=minlp_solver, minlp_solver_args=minlp_args
        )
        self.assertEqual(
            result.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertEqual(
            result.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertIsNone(m.x.value)
        self.assertIsNone(m.d.disjuncts[0].indicator_var.value)
        self.assertIsNone(m.d.disjuncts[1].indicator_var.value)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    def test_LBB_8PP(self):
        """Test the logic-based branch and bound algorithm."""
        exfile = import_file(join(exdir, 'eight_process', 'eight_proc_model.py'))
        eight_process = exfile.build_eight_process_flowsheet()
        results = SolverFactory('gdpopt.lbb').solve(
            eight_process,
            tee=False,
            check_sat=True,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        ct.check_8PP_solution(self, eight_process, results)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    def test_LBB_strip_pack(self):
        """Test logic-based branch and bound with strip packing."""
        exfile = import_file(join(exdir, 'strip_packing', 'strip_packing_concrete.py'))
        strip_pack = exfile.build_rect_strip_packing_model()
        SolverFactory('gdpopt.lbb').solve(
            strip_pack,
            tee=False,
            check_sat=True,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        self.assertTrue(fabs(value(strip_pack.total_length.expr) - 11) <= 1e-2)

    @unittest.skipUnless(license_available, "Problem is too big for unlicensed BARON.")
    @unittest.pytest.mark.expensive
    def test_LBB_constrained_layout(self):
        """Test LBB with constrained layout."""
        exfile = import_file(join(exdir, 'constrained_layout', 'cons_layout_model.py'))
        cons_layout = exfile.build_constrained_layout_model()
        SolverFactory('gdpopt.lbb').solve(
            cons_layout,
            tee=False,
            check_sat=True,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        objective_value = value(cons_layout.min_dist_cost.expr)
        self.assertTrue(
            fabs(objective_value - 41573) <= 200,
            "Objective value of %s instead of 41573" % objective_value,
        )

    def test_LBB_ex_633_trespalacios(self):
        """Test LBB with Francisco thesis example."""
        exfile = import_file(join(exdir, 'small_lit', 'ex_633_trespalacios.py'))
        model = exfile.build_simple_nonconvex_gdp()
        SolverFactory('gdpopt').solve(
            model,
            algorithm='LBB',
            tee=False,
            check_sat=True,
            minlp_solver=minlp_solver,
            minlp_solver_args=minlp_args,
        )
        objective_value = value(model.obj.expr)
        self.assertAlmostEqual(objective_value, 4.46, 2)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr
from pyomo.contrib.piecewise.transform.piecewise_linear_transformation_base import (
    PiecewiseLinearTransformationBase,
)
from pyomo.core import Constraint, NonNegativeIntegers, Suffix, Var
from pyomo.core.base import TransformationFactory
from pyomo.gdp import Disjunct, Disjunction


@TransformationFactory.register(
    'contrib.piecewise.inner_repn_gdp',
    doc="Convert piecewise-linear model to a GDP "
    "using an inner representation of the "
    "simplices that are the domains of the linear "
    "functions.",
)
class InnerRepresentationGDPTransformation(PiecewiseLinearTransformationBase):
    """
    Convert a model involving piecewise linear expressions into a GDP by
    representing the piecewise linear functions as Disjunctions where the
    simplices over which the linear functions are defined are represented
    in an "inner" representation--as convex combinations of their extreme
    points. The multipliers defining the convex combination are local to
    each Disjunct, so there is one per extreme point in each simplex.

    This transformation can be called in one of two ways:
        1) The default, where 'descend_into_expressions' is False. This is
           more computationally efficient, but relies on the
           PiecewiseLinearFunctions being declared on the same Block in which
           they are used in Expressions (if you are hoping to maintain the
           original hierarchical structure of the model). In this mode,
           targets must be Blocks and/or PiecewiseLinearFunctions.
        2) With 'descend_into_expressions' True. This is less computationally
           efficient, but will respect hierarchical structure by finding
           uses of PiecewiseLinearFunctions in Constraint and Objective
           expressions and putting their transformed counterparts on the same
           parent Block as the component owning their parent expression. In
           this mode, targets must be Blocks, Constraints, and/or Objectives.
    """

    CONFIG = PiecewiseLinearTransformationBase.CONFIG()
    _transformation_name = 'pw_linear_inner_repn'

    def _transform_pw_linear_expr(self, pw_expr, pw_linear_func, transformation_block):
        transBlock = transformation_block.transformed_functions[
            len(transformation_block.transformed_functions)
        ]

        # get the PiecewiseLinearFunctionExpression
        dimension = pw_expr.nargs()
        transBlock.disjuncts = Disjunct(NonNegativeIntegers)
        substitute_var = transBlock.substitute_var = Var()
        pw_linear_func.map_transformation_var(pw_expr, substitute_var)
        substitute_var_lb = float('inf')
        substitute_var_ub = -float('inf')
        for simplex, linear_func in zip(
            pw_linear_func._simplices, pw_linear_func._linear_functions
        ):
            disj = transBlock.disjuncts[len(transBlock.disjuncts)]
            disj.lambdas = Var(NonNegativeIntegers, dense=False, bounds=(0, 1))
            extreme_pts = []
            for idx in simplex:
                extreme_pts.append(pw_linear_func._points[idx])

            disj.convex_combo = Constraint(
                expr=sum(disj.lambdas[i] for i in range(len(extreme_pts))) == 1
            )
            linear_func_expr = linear_func(*pw_expr.args)
            disj.set_substitute = Constraint(expr=substitute_var == linear_func_expr)
            lb, ub = compute_bounds_on_expr(linear_func_expr)
            if lb is not None and lb < substitute_var_lb:
                substitute_var_lb = lb
            if ub is not None and ub > substitute_var_ub:
                substitute_var_ub = ub

            @disj.Constraint(range(dimension))
            def linear_combo(disj, i):
                return pw_expr.args[i] == sum(
                    disj.lambdas[j] * pt[i] for j, pt in enumerate(extreme_pts)
                )

            # Mark the lambdas as local so that we don't do anything silly in
            # the hull transformation.
            disj.LocalVars = Suffix(direction=Suffix.LOCAL)
            disj.LocalVars[disj] = [v for v in disj.lambdas.values()]

        if substitute_var_lb < float('inf'):
            transBlock.substitute_var.setlb(substitute_var_lb)
        if substitute_var_ub > -float('inf'):
            transBlock.substitute_var.setub(substitute_var_ub)
        transBlock.pick_a_piece = Disjunction(
            expr=[d for d in transBlock.disjuncts.values()]
        )

        return transBlock.substitute_var



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.errors import InfeasibleConstraintException
import pyomo.common.unittest as unittest
from pyomo.environ import Block, ConcreteModel, Constraint, TransformationFactory
from pyomo.gdp import Disjunct, Disjunction
from pyomo.gdp.util import GDP_Error


class TestTransformCurrentDisjunctiveState(unittest.TestCase):
    def make_two_term_disjunction(self):
        m = ConcreteModel()
        m.d1 = Disjunct()
        m.d2 = Disjunct()
        m.d2.c = Constraint()
        m.disjunction1 = Disjunction(expr=[m.d1, m.d2])

        return m

    def check_fixed_mip(self, m):
        self.assertTrue(m.d1.indicator_var.fixed)
        self.assertTrue(m.d1.active)
        self.assertIs(m.d1.ctype, Block)

        self.assertTrue(m.d2.indicator_var.fixed)
        self.assertFalse(m.d2.active)
        self.assertFalse(m.disjunction1.active)

    def test_fix_disjuncts_and_reverse(self):
        m = self.make_two_term_disjunction()

        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(False)

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m)
        self.check_fixed_mip(m)

        TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(
            m, reverse=reverse
        )
        self.assertFalse(m.d1.indicator_var.fixed)
        self.assertTrue(m.d1.active)
        self.assertTrue(m.d1.indicator_var.value)
        self.assertFalse(m.d2.indicator_var.fixed)
        self.assertTrue(m.d2.active)
        self.assertFalse(m.d2.indicator_var.value)
        self.assertIs(m.d1.ctype, Disjunct)
        self.assertIs(m.d2.ctype, Disjunct)
        self.assertTrue(m.disjunction1.active)

    def test_fix_disjuncts_implied_by_true_disjunct(self):
        m = self.make_two_term_disjunction()

        m.d1.indicator_var.set_value(True)

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m)
        self.check_fixed_mip(m)

        TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(
            m, reverse=reverse
        )
        self.assertFalse(m.d1.indicator_var.fixed)
        self.assertTrue(m.d1.active)
        self.assertTrue(m.d1.indicator_var.value)
        self.assertFalse(m.d2.indicator_var.fixed)
        self.assertTrue(m.d2.active)
        self.assertIsNone(m.d2.indicator_var.value)
        self.assertIs(m.d1.ctype, Disjunct)
        self.assertIs(m.d2.ctype, Disjunct)
        self.assertTrue(m.disjunction1.active)

    def test_fix_disjuncts_implied_by_false_disjunct(self):
        m = self.make_two_term_disjunction()

        m.d2.indicator_var.set_value(False)

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m)
        self.check_fixed_mip(m)

        TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(
            m, reverse=reverse
        )
        self.assertFalse(m.d1.indicator_var.fixed)
        self.assertTrue(m.d1.active)
        self.assertIsNone(m.d1.indicator_var.value)
        self.assertFalse(m.d2.indicator_var.fixed)
        self.assertTrue(m.d2.active)
        self.assertFalse(m.d2.indicator_var.value)
        self.assertIs(m.d1.ctype, Disjunct)
        self.assertIs(m.d2.ctype, Disjunct)
        self.assertTrue(m.disjunction1.active)

    def test_xor_sums_to_0(self):
        m = self.make_two_term_disjunction()
        m.d1.indicator_var.set_value(False)
        m.d2.indicator_var.set_value(False)
        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Logical constraint for Disjunction "
            "'disjunction1' is violated: All the "
            "Disjunct indicator_vars are 'False.'",
        ):
            TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(m)

    def test_xor_sums_to_more_than_1(self):
        m = self.make_two_term_disjunction()
        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(True)
        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Exactly-one constraint for Disjunction "
            "'disjunction1' is violated. The following Disjuncts "
            "are selected: d., d.",
        ):
            TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(m)

    def add_three_term_disjunction(self, m, exactly_one=True):
        m.d = Disjunct([1, 2, 3])
        m.disjunction2 = Disjunction(expr=[m.d[1], m.d[2], m.d[3]], xor=exactly_one)

    def test_ignore_deactivated_disjuncts(self):
        m = self.make_two_term_disjunction()
        self.add_three_term_disjunction(m)

        m.d[1].deactivate()
        m.d[2].indicator_var = True
        m.d[3].indicator_var = False

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m, targets=m.disjunction2)

        self.assertTrue(m.d[1].indicator_var.fixed)
        self.assertFalse(m.d[1].indicator_var.value)
        self.assertFalse(m.d[1].active)

        self.assertTrue(m.d[2].indicator_var.fixed)
        self.assertTrue(m.d[2].indicator_var.value)
        self.assertTrue(m.d[2].active)
        self.assertIs(m.d[2].ctype, Block)

        self.assertTrue(m.d[3].indicator_var.fixed)
        self.assertFalse(m.d[3].indicator_var.value)
        self.assertFalse(m.d[3].active)
        self.assertIs(m.d[3].ctype, Block)

        self.assertFalse(m.disjunction2.active)

        # disjunction1 should be untransformed
        self.assertTrue(m.disjunction1.active)
        self.assertTrue(m.d1.active)
        self.assertTrue(m.d2.active)
        self.assertIs(m.d1.ctype, Disjunct)
        self.assertIs(m.d2.ctype, Disjunct)
        self.assertIsNone(m.d1.indicator_var.value)
        self.assertIsNone(m.d2.indicator_var.value)

        TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(
            m, reverse=reverse
        )

        self.assertTrue(m.d[1].indicator_var.fixed)
        self.assertFalse(m.d[1].indicator_var.value)
        self.assertFalse(m.d[1].active)
        self.assertIs(m.d[1].ctype, Disjunct)

        self.assertFalse(m.d[2].indicator_var.fixed)
        self.assertTrue(m.d[2].indicator_var.value)
        self.assertTrue(m.d[2].active)
        self.assertIs(m.d[2].ctype, Disjunct)

        self.assertFalse(m.d[3].indicator_var.fixed)
        self.assertFalse(m.d[3].indicator_var.value)
        self.assertTrue(m.d[3].active)
        self.assertIs(m.d[3].ctype, Disjunct)

        self.assertTrue(m.disjunction2.active)

    def test_no_partial_reverse(self):
        m = self.make_two_term_disjunction()
        self.add_three_term_disjunction(m)

        m.d1.indicator_var.set_value(True)
        m.d2.indicator_var.set_value(False)
        m.d[2].indicator_var = True
        m.d[3].indicator_var = False

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m)

        self.check_fixed_mip(m)

        with self.assertRaisesRegex(
            ValueError,
            "The 'gdp.transform_current_disjunctive_state' transformation "
            "cannot be called with both targets and a reverse token "
            "specified. If reversing the transformation, do not include "
            "targets: The reverse transformation will restore all the "
            "components the original transformation call transformed.",
        ):
            TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(
                m, reverse=reverse, targets=m.disjunction2
            )

    def test_at_least_one_disjunction(self):
        m = ConcreteModel()
        self.add_three_term_disjunction(m, exactly_one=False)

        m.d[1].indicator_var.fix(True)
        m.d[2].indicator_var = True
        m.d[3].indicator_var = False

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m)

        self.assertTrue(m.d[1].indicator_var.fixed)
        self.assertTrue(m.d[1].indicator_var.value)
        self.assertTrue(m.d[1].active)
        self.assertIs(m.d[1].ctype, Block)

        self.assertTrue(m.d[2].indicator_var.fixed)
        self.assertTrue(m.d[2].indicator_var.value)
        self.assertTrue(m.d[2].active)
        self.assertIs(m.d[2].ctype, Block)

        self.assertTrue(m.d[3].indicator_var.fixed)
        self.assertFalse(m.d[3].indicator_var.value)
        self.assertFalse(m.d[3].active)
        self.assertIs(m.d[3].ctype, Block)

        self.assertFalse(m.disjunction2.active)

        reverse = TransformationFactory(
            'gdp.transform_current_disjunctive_state'
        ).apply_to(m, reverse=reverse)

        self.assertTrue(m.d[1].indicator_var.fixed)
        self.assertTrue(m.d[1].indicator_var.value)
        self.assertTrue(m.d[1].active)
        self.assertIs(m.d[1].ctype, Disjunct)

        self.assertFalse(m.d[2].indicator_var.fixed)
        self.assertTrue(m.d[2].indicator_var.value)
        self.assertTrue(m.d[2].active)
        self.assertIs(m.d[2].ctype, Disjunct)

        self.assertFalse(m.d[3].indicator_var.fixed)
        self.assertFalse(m.d[3].indicator_var.value)
        self.assertTrue(m.d[3].active)
        self.assertIs(m.d[3].ctype, Disjunct)

        self.assertTrue(m.disjunction2.active)

    def test_at_least_one_disjunction_infeasible(self):
        m = ConcreteModel()
        self.add_three_term_disjunction(m, exactly_one=False)

        m.d[1].indicator_var.fix(False)
        m.d[2].indicator_var = False
        m.d[3].indicator_var = False

        with self.assertRaisesRegex(
            InfeasibleConstraintException,
            "Logical constraint for Disjunction "
            "'disjunction2' is violated: All "
            "the Disjunct indicator_vars are 'False.'",
        ):
            TransformationFactory('gdp.transform_current_disjunctive_state').apply_to(m)

    def test_not_enough_info_to_fully_transform(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2, 3, 4])
        m.disj1 = Disjunction(expr=[m.d[1], m.d[2]])
        m.disj2 = Disjunction(expr=[m.d[3], m.d[4]])

        m.d[1].indicator_var = True

        with self.assertRaisesRegex(
            GDP_Error,
            "Disjunction 'disj2' does not contain enough Disjuncts with "
            "values in their indicator_vars to specify which Disjuncts "
            "are True. Cannot fully transform model.",
        ):
            reverse = TransformationFactory(
                'gdp.transform_current_disjunctive_state'
            ).apply_to(m)

    def test_not_enough_info_in_single_disjunction_to_fully_transform_xor(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2, 3, 4])
        m.disj1 = Disjunction(expr=[m.d[1], m.d[2], m.d[3], m.d[4]])
        m.d[1].indicator_var = False
        m.d[2].indicator_var = False

        with self.assertRaisesRegex(
            GDP_Error,
            "Disjunction 'disj1' does not contain enough Disjuncts with "
            "values in their indicator_vars to specify which Disjuncts "
            "are True. Cannot fully transform model.",
        ):
            reverse = TransformationFactory(
                'gdp.transform_current_disjunctive_state'
            ).apply_to(m)

    def test_not_enough_info_in_single_disjunction_to_fully_transform_or(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2, 3, 4])
        m.disj1 = Disjunction(expr=[m.d[1], m.d[2], m.d[3], m.d[4]], xor=False)
        m.d[1].indicator_var = True
        m.d[2].indicator_var = False

        with self.assertRaisesRegex(
            GDP_Error,
            "Disjunction 'disj1' does not contain enough Disjuncts with "
            "values in their indicator_vars to specify which Disjuncts "
            "are True. Cannot fully transform model.",
        ):
            reverse = TransformationFactory(
                'gdp.transform_current_disjunctive_state'
            ).apply_to(m)

    def test_complain_about_dangling_disjuncts(self):
        m = ConcreteModel()
        m.d = Disjunct([1, 2, 3, 4])
        m.disj1 = Disjunction(expr=[m.d[1], m.d[2], m.d[3]])
        m.d[1].indicator_var = True

        with self.assertRaisesRegex(
            GDP_Error,
            r"Found active Disjuncts on the model that "
            r"were not included in any Disjunctions:\nd\[4\]\nPlease "
            r"deactivate them or include them in a Disjunction.",
        ):
            reverse = TransformationFactory(
                'gdp.transform_current_disjunctive_state'
            ).apply_to(m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import numpy as np
from scipy.sparse import coo_matrix
import pyomo.environ as pyo
from pyomo.contrib.pynumero.interfaces.external_grey_box import (
    ExternalGreyBoxModel,
    ExternalGreyBoxBlock,
)


class Unconstrained(ExternalGreyBoxModel):
    """
    min (x+2)**2 + (y-2)**2
    """

    def input_names(self):
        return ['x', 'y']

    def set_input_values(self, input_values):
        self._input_values = list(input_values)

    def has_objective(self):
        return True

    def evaluate_objective(self):
        x = self._input_values[0]
        y = self._input_values[1]
        return (x + 2) ** 2 + (y - 2) ** 2

    def evaluate_grad_objective(self):
        x = self._input_values[0]
        y = self._input_values[1]
        return np.asarray([2 * (x + 2), 2 * (y - 2)], dtype=float)


class Constrained(ExternalGreyBoxModel):
    """
    min x**2 + y**2
    s.t. 0 == y - exp(x)
    """

    def input_names(self):
        return ['x', 'y']

    def set_input_values(self, input_values):
        self._input_values = list(input_values)

    def has_objective(self):
        return True

    def evaluate_objective(self):
        x = self._input_values[0]
        y = self._input_values[1]
        return x**2 + y**2

    def evaluate_grad_objective(self):
        x = self._input_values[0]
        y = self._input_values[1]
        return np.asarray([2 * x, 2 * y], dtype=float)

    def equality_constraint_names(self):
        return ['c1']

    def evaluate_equality_constraints(self):
        x = self._input_values[0]
        y = self._input_values[1]
        return np.asarray([y - math.exp(x)], dtype=float)

    def evaluate_jacobian_equality_constraints(self):
        x = self._input_values[0]
        row = [0, 0]
        col = [0, 1]
        data = [-math.exp(x), 1]
        jac = coo_matrix((data, (row, col)), shape=(1, 2))
        return jac


class ConstrainedWithHessian(Constrained):
    def evaluate_hessian_objective(self):
        row = [0, 1]
        col = [0, 1]
        data = [2, 2]
        hess = coo_matrix((data, (row, col)), shape=(2, 2))
        return hess

    def set_equality_constraint_multipliers(self, eq_con_multiplier_values):
        self._dual = eq_con_multiplier_values[0]

    def evaluate_hessian_equality_constraints(self):
        x = self._input_values[0]
        row = [0]
        col = [0]
        data = [-math.exp(x) * self._dual]
        hess = coo_matrix((data, (row, col)), shape=(2, 2))
        return hess


def solve_unconstrained():
    m = pyo.ConcreteModel()
    m.z = pyo.Var()
    m.grey_box = ExternalGreyBoxBlock(external_model=Unconstrained())
    m.c = pyo.Constraint(expr=m.z == m.grey_box.inputs['x'] + 1)

    opt = pyo.SolverFactory('cyipopt')
    opt.config.options['hessian_approximation'] = 'limited-memory'
    res = opt.solve(m, tee=True)
    pyo.assert_optimal_termination(res)
    x = m.grey_box.inputs['x'].value
    y = m.grey_box.inputs['y'].value
    assert math.isclose(x, -2)
    assert math.isclose(y, 2)
    return m


def solve_constrained():
    m = pyo.ConcreteModel()
    m.z = pyo.Var()
    m.grey_box = ExternalGreyBoxBlock(external_model=Constrained())
    m.c2 = pyo.Constraint(expr=m.z == m.grey_box.inputs['x'] + 1)

    opt = pyo.SolverFactory('cyipopt')
    opt.config.options['hessian_approximation'] = 'limited-memory'
    res = opt.solve(m, tee=True)
    pyo.assert_optimal_termination(res)
    x = m.grey_box.inputs['x'].value
    y = m.grey_box.inputs['y'].value
    assert math.isclose(x, -0.4263027509962655)
    assert math.isclose(y, 0.6529186403960969)
    return m


def solve_constrained_with_hessian():
    m = pyo.ConcreteModel()
    m.z = pyo.Var()
    m.grey_box = ExternalGreyBoxBlock(external_model=ConstrainedWithHessian())
    m.c2 = pyo.Constraint(expr=m.z == m.grey_box.inputs['x'] + 1)

    opt = pyo.SolverFactory('cyipopt')
    res = opt.solve(m, tee=True)
    pyo.assert_optimal_termination(res)
    x = m.grey_box.inputs['x'].value
    y = m.grey_box.inputs['y'].value
    assert math.isclose(x, -0.4263027509962655)
    assert math.isclose(y, 0.6529186403960969)
    return m


if __name__ == '__main__':
    m = solve_constrained_with_hessian()
    print(f"x: {m.grey_box.inputs['x'].value}")
    print(f"y: {m.grey_box.inputs['y'].value}")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
import logging
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    In,
    PositiveFloat,
    PositiveInt,
    NonNegativeInt,
)
from pyomo.contrib.gdpopt.util import _DoNothing, a_logger
from pyomo.common.deprecation import deprecation_warning

_supported_algorithms = {
    'OA': ('mindtpy.oa', 'Outer Approximation'),
    'ECP': ('mindtpy.ecp', 'Extended Cutting Plane'),
    'GOA': ('mindtpy.goa', 'Global Outer Approximation'),
    'FP': ('mindtpy.fp', 'Feasibility Pump'),
}


def _get_MindtPy_config():
    """Set up the configurations for MindtPy.

    Returns
    -------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy
    """
    CONFIG = ConfigBlock('MindtPy')

    _add_common_configs(CONFIG)
    _add_subsolver_configs(CONFIG)
    _add_tolerance_configs(CONFIG)
    _add_fp_configs(CONFIG)
    _add_bound_configs(CONFIG)
    _add_roa_configs(CONFIG)
    return CONFIG


def _get_MindtPy_OA_config():
    """Set up the configurations for MindtPy-OA.

    Returns
    -------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy
    """
    CONFIG = ConfigBlock('MindtPy-OA')

    _add_common_configs(CONFIG)
    _add_oa_configs(CONFIG)
    _add_roa_configs(CONFIG)
    _add_fp_configs(CONFIG)
    _add_oa_cuts_configs(CONFIG)
    _add_subsolver_configs(CONFIG)
    _add_tolerance_configs(CONFIG)
    _add_bound_configs(CONFIG)
    return CONFIG


def _get_MindtPy_GOA_config():
    """Set up the configurations for MindtPy-GOA.

    Returns
    -------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy
    """
    CONFIG = ConfigBlock('MindtPy-GOA')

    _add_common_configs(CONFIG)
    _add_goa_configs(CONFIG)
    _add_oa_cuts_configs(CONFIG)
    _add_subsolver_configs(CONFIG)
    _add_tolerance_configs(CONFIG)
    _add_bound_configs(CONFIG)
    return CONFIG


def _get_MindtPy_ECP_config():
    """Set up the configurations for MindtPy-ECP.

    Returns
    -------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy
    """
    CONFIG = ConfigBlock('MindtPy-GOA')

    _add_common_configs(CONFIG)
    _add_ecp_configs(CONFIG)
    _add_oa_cuts_configs(CONFIG)
    _add_subsolver_configs(CONFIG)
    _add_tolerance_configs(CONFIG)
    _add_bound_configs(CONFIG)
    return CONFIG


def _get_MindtPy_FP_config():
    """Set up the configurations for MindtPy-FP.

    Returns
    -------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy
    """
    CONFIG = ConfigBlock('MindtPy-GOA')
    CONFIG.declare(
        'init_strategy',
        ConfigValue(
            default='FP',
            domain=In(['FP']),
            description='Initialization strategy',
            doc='Initialization strategy used by any method. Currently the '
            'continuous relaxation of the MINLP (rNLP), solve a maximal '
            'covering problem (max_binary), and fix the initial value for '
            'the integer variables (initial_binary).',
        ),
    )

    _add_common_configs(CONFIG)
    _add_fp_configs(CONFIG)
    _add_oa_cuts_configs(CONFIG)
    _add_subsolver_configs(CONFIG)
    _add_tolerance_configs(CONFIG)
    _add_bound_configs(CONFIG)
    return CONFIG


def _add_oa_configs(CONFIG):
    CONFIG.declare(
        'heuristic_nonconvex',
        ConfigValue(
            default=False,
            description='Use dual solution from the NLP solver and slack variables to add OA cuts for equality constraints (Equality relaxation)'
            'and minimize the sum of the slack variables (Augmented Penalty).',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'init_strategy',
        ConfigValue(
            default='rNLP',
            domain=In(['rNLP', 'initial_binary', 'max_binary', 'FP']),
            description='Initialization strategy',
            doc='Initialization strategy used by any method. Currently the '
            'continuous relaxation of the MINLP (rNLP), solve a maximal '
            'covering problem (max_binary), and fix the initial value for '
            'the integer variables (initial_binary).',
        ),
    )


def _add_oa_cuts_configs(CONFIG):
    CONFIG.declare(
        'add_slack',
        ConfigValue(
            default=False,
            description='Whether add slack variable here.'
            'slack variables here are used to deal with nonconvex MINLP.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'max_slack',
        ConfigValue(
            default=1000.0,
            domain=PositiveFloat,
            description='Maximum slack variable',
            doc='Maximum slack variable value allowed for the Outer Approximation '
            'cuts.',
        ),
    )
    CONFIG.declare(
        'OA_penalty_factor',
        ConfigValue(
            default=1000.0,
            domain=PositiveFloat,
            description='Outer Approximation slack penalty factor',
            doc='In the objective function of the Outer Approximation method, the '
            'slack variables corresponding to all the constraints get '
            'multiplied by this number and added to the objective.',
        ),
    )
    CONFIG.declare(
        'equality_relaxation',
        ConfigValue(
            default=False,
            description='Use dual solution from the NLP solver to add OA cuts for equality constraints.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'linearize_inactive',
        ConfigValue(
            default=False,
            description='Add OA cuts for inactive constraints.',
            domain=bool,
        ),
    )


def _add_goa_configs(CONFIG):
    CONFIG.declare(
        'init_strategy',
        ConfigValue(
            default='rNLP',
            domain=In(['rNLP', 'initial_binary', 'max_binary']),
            description='Initialization strategy',
            doc='Initialization strategy used by any method. Currently the '
            'continuous relaxation of the MINLP (rNLP), solve a maximal '
            'covering problem (max_binary), and fix the initial value for '
            'the integer variables (initial_binary).',
        ),
    )


def _add_ecp_configs(CONFIG):
    CONFIG.declare(
        'ecp_tolerance',
        ConfigValue(
            default=None,
            domain=PositiveFloat,
            description='ECP tolerance',
            doc='Feasibility tolerance used to determine the stopping criterion in'
            'the ECP method. As long as nonlinear constraint are violated for '
            'more than this tolerance, the method will keep iterating.',
        ),
    )
    CONFIG.declare(
        'init_strategy',
        ConfigValue(
            default='max_binary',
            domain=In(['rNLP', 'max_binary', 'FP']),
            description='Initialization strategy',
            doc='Initialization strategy used by any method. Currently the '
            'continuous relaxation of the MINLP (rNLP), solve a maximal '
            'covering problem (max_binary), and fix the initial value for '
            'the integer variables (initial_binary).',
        ),
    )


def _add_common_configs(CONFIG):
    CONFIG.declare(
        'iteration_limit',
        ConfigValue(
            default=50,
            domain=NonNegativeInt,
            description='Iteration limit',
            doc='Number of maximum iterations in the decomposition methods.',
        ),
    )
    CONFIG.declare(
        'stalling_limit',
        ConfigValue(
            default=15,
            domain=PositiveInt,
            description='Stalling limit',
            doc='Stalling limit for primal bound progress in the decomposition methods.',
        ),
    )
    CONFIG.declare(
        'time_limit',
        ConfigValue(
            default=600,
            domain=PositiveInt,
            description='Time limit (seconds, default=600)',
            doc='Seconds allowed until terminated. Note that the time limit can'
            'currently only be enforced between subsolver invocations. You may'
            'need to set subsolver time limits as well.',
        ),
    )
    CONFIG.declare(
        'strategy',
        ConfigValue(
            default='OA',
            domain=In(['OA', 'ECP', 'GOA', 'FP']),
            description='Decomposition strategy',
            doc='MINLP Decomposition strategy to be applied to the method. '
            'Currently available Outer Approximation (OA), Extended Cutting '
            'Plane (ECP), Global Outer Approximation (GOA) and Feasibility Pump (FP).',
        ),
    )
    CONFIG.declare(
        'add_regularization',
        ConfigValue(
            default=None,
            domain=In(
                [
                    'level_L1',
                    'level_L2',
                    'level_L_infinity',
                    'grad_lag',
                    'hess_lag',
                    'hess_only_lag',
                    'sqp_lag',
                ]
            ),
            description='add regularization',
            doc='Solving a regularization problem before solve the fixed subproblem'
            'the objective function of the regularization problem.',
        ),
    )
    CONFIG.declare(
        'call_after_main_solve',
        ConfigValue(
            default=_DoNothing(),
            domain=None,
            description='Function to be executed after every main problem',
            doc='Callback hook after a solution of the main problem.',
        ),
    )
    CONFIG.declare(
        'call_before_subproblem_solve',
        ConfigValue(
            default=_DoNothing(),
            domain=None,
            description='Function to be executed before every subproblem',
            doc='Callback hook before a solution of the nonlinear subproblem.',
        ),
    )
    CONFIG.declare(
        'call_after_subproblem_solve',
        ConfigValue(
            default=_DoNothing(),
            domain=None,
            description='Function to be executed after every subproblem',
            doc='Callback hook after a solution of the nonlinear subproblem.',
        ),
    )
    CONFIG.declare(
        'call_after_subproblem_feasible',
        ConfigValue(
            default=_DoNothing(),
            domain=None,
            description='Function to be executed after every feasible subproblem',
            doc='Callback hook after a feasible solution'
            ' of the nonlinear subproblem.',
        ),
    )
    CONFIG.declare(
        'tee',
        ConfigValue(
            default=False, description='Stream output to terminal.', domain=bool
        ),
    )
    CONFIG.declare(
        'logger',
        ConfigValue(
            default='pyomo.contrib.mindtpy',
            description='The logger object or name to use for reporting.',
            domain=a_logger,
        ),
    )
    CONFIG.declare(
        'logging_level',
        ConfigValue(
            default=logging.INFO,
            domain=NonNegativeInt,
            description='The logging level for MindtPy.'
            'CRITICAL = 50, ERROR = 40, WARNING = 30, INFO = 20, DEBUG = 10, NOTSET = 0',
        ),
    )
    CONFIG.declare(
        'integer_to_binary',
        ConfigValue(
            default=False,
            description='Convert integer variables to binaries (for no-good cuts).',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'add_no_good_cuts',
        ConfigValue(
            default=False,
            description='Add no-good cuts (no-good cuts) to binary variables to disallow same integer solution again.'
            'Note that integer_to_binary flag needs to be used to apply it to actual integers and not just binaries.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'use_tabu_list',
        ConfigValue(
            default=False,
            description='Use tabu list and incumbent callback to disallow same integer solution again.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'single_tree',
        ConfigValue(
            default=False,
            description='Use single tree implementation in solving the MIP main problem.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'solution_pool',
        ConfigValue(
            default=False,
            description='Use solution pool in solving the MIP main problem.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'num_solution_iteration',
        ConfigValue(
            default=5,
            description='The number of MIP solutions (from the solution pool) used to generate the fixed NLP subproblem in each iteration.',
            domain=PositiveInt,
        ),
    )
    CONFIG.declare(
        'cycling_check',
        ConfigValue(
            default=True,
            description='Check if OA algorithm is stalled in a cycle and terminate.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'feasibility_norm',
        ConfigValue(
            default='L_infinity',
            domain=In(['L1', 'L2', 'L_infinity']),
            description='Different forms of objective function in feasibility subproblem.',
        ),
    )
    CONFIG.declare(
        'differentiate_mode',
        ConfigValue(
            default='reverse_symbolic',
            domain=In(['reverse_symbolic', 'sympy']),
            description='Differentiate mode to calculate jacobian.',
        ),
    )
    CONFIG.declare(
        'use_mcpp',
        ConfigValue(
            default=False,
            description="Use package MC++ to set a bound for variable 'objective_value', which is introduced when the original problem's objective function is nonlinear.",
            domain=bool,
        ),
    )
    CONFIG.declare(
        'calculate_dual_at_solution',
        ConfigValue(
            default=False,
            description='Calculate duals of the NLP subproblem.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'use_fbbt',
        ConfigValue(
            default=False,
            description='Use fbbt to tighten the feasible region of the problem.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'use_dual_bound',
        ConfigValue(
            default=True,
            description='Add dual bound constraint to enforce the objective satisfies best-found dual bound.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'partition_obj_nonlinear_terms',
        ConfigValue(
            default=True,
            description='Partition objective with the sum of nonlinear terms using epigraph reformulation.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'quadratic_strategy',
        ConfigValue(
            default=0,
            domain=In([0, 1, 2]),
            description='How to treat the quadratic terms in MINLP.'
            '0 : treat as nonlinear terms'
            '1 : only use quadratic terms in objective function directly in main problem'
            '2 : use quadratic terms in objective function and constraints in main problem',
        ),
    )
    CONFIG.declare(
        'move_objective',
        ConfigValue(
            default=False,
            domain=bool,
            description='Whether to replace the objective function to constraint using epigraph constraint.',
        ),
    )
    CONFIG.declare(
        'add_cuts_at_incumbent',
        ConfigValue(
            default=False,
            description='Whether to add lazy cuts to the main problem at the incumbent solution found in the branch & bound tree',
            domain=bool,
        ),
    )


def _add_subsolver_configs(CONFIG):
    """Adds the subsolver-related configurations.

    Parameters
    ----------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy.
    """
    CONFIG.declare(
        'nlp_solver',
        ConfigValue(
            default='ipopt',
            domain=In(['ipopt', 'appsi_ipopt', 'gams', 'baron', 'cyipopt']),
            description='NLP subsolver name',
            doc='Which NLP subsolver is going to be used for solving the nonlinear'
            'subproblems.',
        ),
    )
    CONFIG.declare(
        'nlp_solver_args',
        ConfigBlock(
            implicit=True,
            description='NLP subsolver options',
            doc='Which NLP subsolver options to be passed to the solver while '
            'solving the nonlinear subproblems.',
        ),
    )
    CONFIG.declare(
        'mip_solver',
        ConfigValue(
            default='glpk',
            domain=In(
                [
                    'gurobi',
                    'cplex',
                    'cbc',
                    'glpk',
                    'gams',
                    'gurobi_persistent',
                    'cplex_persistent',
                    'appsi_cplex',
                    'appsi_gurobi',
                    'appsi_highs',
                ]
            ),
            description='MIP subsolver name',
            doc='Which MIP subsolver is going to be used for solving the mixed-'
            'integer main problems.',
        ),
    )
    CONFIG.declare(
        'mip_solver_args',
        ConfigBlock(
            implicit=True,
            description='MIP subsolver options',
            doc='Which MIP subsolver options to be passed to the solver while '
            'solving the mixed-integer main problems.',
        ),
    )
    CONFIG.declare(
        'mip_solver_mipgap',
        ConfigValue(
            default=1e-4,
            domain=PositiveFloat,
            description='Mipgap passed to MIP solver.',
        ),
    )
    CONFIG.declare(
        'threads',
        ConfigValue(
            default=0,
            domain=NonNegativeInt,
            description='Threads',
            doc='Threads used by MIP solver and NLP solver.',
        ),
    )
    CONFIG.declare(
        'regularization_mip_threads',
        ConfigValue(
            default=0,
            domain=NonNegativeInt,
            description='regularization MIP threads',
            doc='Threads used by MIP solver to solve regularization main problem.',
        ),
    )
    CONFIG.declare(
        'solver_tee',
        ConfigValue(
            default=False,
            description='Stream the output of MIP solver and NLP solver to terminal.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'mip_solver_tee',
        ConfigValue(
            default=False,
            description='Stream the output of MIP solver to terminal.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'nlp_solver_tee',
        ConfigValue(
            default=False,
            description='Stream the output of nlp solver to terminal.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'mip_regularization_solver',
        ConfigValue(
            default=None,
            domain=In(
                [
                    'gurobi',
                    'cplex',
                    'cbc',
                    'glpk',
                    'gams',
                    'gurobi_persistent',
                    'cplex_persistent',
                    'appsi_cplex',
                    'appsi_gurobi',
                    'appsi_highs',
                ]
            ),
            description='MIP subsolver for regularization problem',
            doc='Which MIP subsolver is going to be used for solving the regularization problem.',
        ),
    )


def _add_tolerance_configs(CONFIG):
    """Adds the tolerance-related configurations.

    Parameters
    ----------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy.
    """
    CONFIG.declare(
        'absolute_bound_tolerance',
        ConfigValue(
            default=1e-4,
            domain=PositiveFloat,
            description='Bound tolerance',
            doc='Absolute tolerance for bound feasibility checks.',
        ),
    )
    CONFIG.declare(
        'relative_bound_tolerance',
        ConfigValue(
            default=1e-3,
            domain=PositiveFloat,
            description='Relative bound tolerance',
            doc='Relative tolerance for bound feasibility checks. '
            ':math:`|Primal Bound - Dual Bound| / (1e-10 + |Primal Bound|) <= relative tolerance`',
        ),
    )
    CONFIG.declare(
        'small_dual_tolerance',
        ConfigValue(
            default=1e-8,
            description='When generating cuts, small duals multiplied '
            'by expressions can cause problems. Exclude all duals '
            'smaller in absolute value than the following.',
        ),
    )
    CONFIG.declare(
        'integer_tolerance',
        ConfigValue(default=1e-5, description='Tolerance on integral values.'),
    )
    CONFIG.declare(
        'constraint_tolerance',
        ConfigValue(default=1e-6, description='Tolerance on constraint satisfaction.'),
    )
    CONFIG.declare(
        'variable_tolerance',
        ConfigValue(default=1e-8, description='Tolerance on variable bounds.'),
    )
    CONFIG.declare(
        'zero_tolerance',
        ConfigValue(default=1e-8, description='Tolerance on variable equal to zero.'),
    )


def _add_bound_configs(CONFIG):
    """Adds the bound-related configurations.

    Parameters
    ----------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy.
    """
    CONFIG.declare(
        'obj_bound',
        ConfigValue(
            default=1e15,
            domain=PositiveFloat,
            description='Bound applied to the linearization of the objective function if main MIP is unbounded.',
        ),
    )
    CONFIG.declare(
        'continuous_var_bound',
        ConfigValue(
            default=1e10,
            description='Default bound added to unbounded continuous variables in nonlinear constraint if single tree is activated.',
            domain=PositiveFloat,
        ),
    )
    CONFIG.declare(
        'integer_var_bound',
        ConfigValue(
            default=1e9,
            description='Default bound added to unbounded integral variables in nonlinear constraint if single tree is activated.',
            domain=PositiveFloat,
        ),
    )
    CONFIG.declare(
        'initial_bound_coef',
        ConfigValue(
            default=1e-1,
            domain=PositiveFloat,
            description='The coefficient used to approximate the initial primal/dual bound.',
        ),
    )


def _add_fp_configs(CONFIG):
    """Adds the feasibility pump-related configurations.

    Parameters
    ----------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy.
    """
    CONFIG.declare(
        'fp_cutoffdecr',
        ConfigValue(
            default=1e-1,
            domain=PositiveFloat,
            description='Additional relative decrement of cutoff value for the original objective function.',
        ),
    )
    CONFIG.declare(
        'fp_iteration_limit',
        ConfigValue(
            default=20,
            domain=PositiveInt,
            description='Feasibility pump iteration limit',
            doc='Number of maximum iterations in the feasibility pump methods.',
        ),
    )
    # TODO: integrate this option
    CONFIG.declare(
        'fp_projcuts',
        ConfigValue(
            default=True,
            description='Whether to add cut derived from regularization of MIP solution onto NLP feasible set.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'fp_transfercuts',
        ConfigValue(
            default=True,
            description='Whether to transfer cuts from the Feasibility Pump MIP to main MIP in selected strategy (all except from the round in which the FP MIP became infeasible).',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'fp_projzerotol',
        ConfigValue(
            default=1e-4,
            domain=PositiveFloat,
            description='Tolerance on when to consider optimal value of regularization problem as zero, which may trigger the solution of a Sub-NLP.',
        ),
    )
    CONFIG.declare(
        'fp_mipgap',
        ConfigValue(
            default=1e-2,
            domain=PositiveFloat,
            description='Optimality tolerance (relative gap) to use for solving MIP regularization problem.',
        ),
    )
    CONFIG.declare(
        'fp_discrete_only',
        ConfigValue(
            default=True,
            description='Only calculate the distance among discrete variables in regularization problems.',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'fp_main_norm',
        ConfigValue(
            default='L1',
            domain=In(['L1', 'L2', 'L_infinity']),
            description='Different forms of objective function MIP regularization problem.',
        ),
    )
    CONFIG.declare(
        'fp_norm_constraint',
        ConfigValue(
            default=True,
            description='Whether to add the norm constraint to FP-NLP',
            domain=bool,
        ),
    )
    CONFIG.declare(
        'fp_norm_constraint_coef',
        ConfigValue(
            default=1,
            domain=PositiveFloat,
            description='The coefficient in the norm constraint, correspond to the Beta in the paper.',
        ),
    )


def _add_roa_configs(CONFIG):
    """Adds the ROA-related configurations.

    Parameters
    ----------
    CONFIG : ConfigBlock
        The specific configurations for MindtPy.
    """
    CONFIG.declare(
        'level_coef',
        ConfigValue(
            default=0.5,
            domain=PositiveFloat,
            description='The coefficient in the regularization main problem'
            'represents how much the linear approximation of the MINLP problem is trusted.',
        ),
    )
    CONFIG.declare(
        'solution_limit',
        ConfigValue(
            default=10,
            domain=PositiveInt,
            description='The solution limit for the regularization problem since it does not need to be solved to optimality.',
        ),
    )
    CONFIG.declare(
        'sqp_lag_scaling_coef',
        ConfigValue(
            default='fixed',
            domain=In(['fixed', 'variable_dependent']),
            description='The coefficient used to scale the L2 norm in sqp_lag.',
        ),
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test the Pyomo BAR writer
#

import re
import glob
import os
from os.path import abspath, dirname, join

from filecmp import cmp
import pyomo.common.unittest as unittest
from pyomo.common.dependencies import attempt_import
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

import pyomo.scripting.pyomo_main as main

parameterized, param_available = attempt_import('parameterized')
if not param_available:
    raise unittest.SkipTest('Parameterized is not available.')

currdir = this_file_dir()
datadir = abspath(join(currdir, "..", "ampl"))
deleteFiles = True

# add test methods to classes
invalidlist = []
validlist = []

invalid_tests = {'small14'}
for f in glob.glob(join(datadir, '*_testCase.py')):
    name = re.split('[._]', os.path.basename(f))[0]
    if name in invalid_tests:
        # Create some list
        invalidlist.append((name, datadir))
    else:
        validlist.append((name, datadir))

for f in glob.glob(join(currdir, '*_testCase.py')):
    name = re.split('[._]', os.path.basename(f))[0]
    validlist.append((name, currdir))


class Tests(unittest.TestCase):
    def pyomo(self, cmd):
        os.chdir(currdir)
        output = main.main(['convert', '--logging=quiet', '-c'] + cmd)
        return output

    def setUp(self):
        TempfileManager.push()

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())


class BaselineTests(Tests):
    #
    # The following test generates a GMS file for the test case
    # and checks that it matches the current pyomo baseline GMS file
    #

    @parameterized.parameterized.expand(input=validlist)
    def gams_writer_baseline_test(self, name, targetdir):
        baseline = join(currdir, name + '.pyomo.gms')
        testFile = TempfileManager.create_tempfile(suffix=name + '.test.gms')
        cmd = ['--output=' + testFile, join(targetdir, name + '_testCase.py')]
        if os.path.exists(join(targetdir, name + '.dat')):
            cmd.append(join(targetdir, name + '.dat'))
        self.pyomo(cmd)

        # Check that the pyomo nl file matches its own baseline
        try:
            self.assertTrue(cmp(testFile, baseline))
        except:
            with open(baseline, 'r') as f1, open(testFile, 'r') as f2:
                f1_contents = list(filter(None, f1.read().split()))
                f2_contents = list(filter(None, f2.read().split()))
                self.assertEqual(
                    f1_contents,
                    f2_contents,
                    "\n\nbaseline: %s\ntestFile: %s\n" % (baseline, testFile),
                )

    @parameterized.parameterized.expand(input=invalidlist)
    def gams_writer_test_invalid(self, name, targetdir):
        with self.assertRaisesRegex(
            RuntimeError, "GAMS files cannot represent the unary function"
        ):
            testFile = TempfileManager.create_tempfile(suffix=name + '.test.gms')
            cmd = ['--output=' + testFile, join(targetdir, name + '_testCase.py')]
            if os.path.exists(join(targetdir, name + '.dat')):
                cmd.append(join(targetdir, name + '.dat'))
            self.pyomo(cmd)


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()
